import argparse
import copy
import traceback
from io import TextIOWrapper

import ophyd
import yaml
from bec_lib.scibec_validator import SciBecValidator

from ophyd_devices.utils.bec_device_base import BECDevice

try:
    from bec_plugins import devices as plugin_devices
except ImportError:
    plugin_devices = None

try:
    from device_server.devices.devicemanager import DeviceManagerDS as device_manager
except ImportError:
    device_manager = None


class StaticDeviceTest:
    """Class to perform tests on an ophyd device config file."""

    def __init__(self, config: str, output_file: TextIOWrapper) -> None:
        """
        Args:
            config(str): path to the config file
            output_file(TextIOWrapper): file to write the output to
        """
        self.config = self.read_config(config)
        self.file = output_file

    @staticmethod
    def read_config(config) -> dict:
        """
        Read the config file

        Args:
            config(str): path to the config file

        Returns:
            dict: config content
        """
        content = None
        with open(config, "r", encoding="utf-8") as file:
            file_content = file.read()
            content = yaml.safe_load(file_content)
        return content

    def _check_all_signals_of_device(self, name: str, device: ophyd.Device) -> None:
        """
        Check if all signals of the device that are not omitted are configured with auto_monitor=True

        Args:
            name(str): name of the device
            device(ophyd.Device): device object
        """
        for _, sub_name, item in device.walk_components():
            if not issubclass(item.cls, ophyd.signal.EpicsSignalBase):
                continue
            if not item.is_signal:
                continue
            if not item.kind == ophyd.Kind.omitted:
                continue
            # check if auto_monitor is in kwargs
            self._has_auto_monitor(f"{name}/{sub_name}", item.kwargs)

    def _check_epics_motor(self, name: str, config: dict) -> None:
        """
        Check if the epics motor config is valid.

        Args:
            name(str): name of the device
            config(dict): device config
        """
        if "prefix" in config["deviceConfig"]:
            return
        msg_suffix = ""

        # check if the device specifies a read_pv instead of a prefix.
        # This is a common copy-paste error.
        if "read_pv" in config["deviceConfig"]:
            msg_suffix = "Maybe a typo? The device specifies a read_pv instead."
        raise ValueError(f"{name}: does not specify the prefix. {msg_suffix}")

    def _check_epics_signal(self, name: str, config: dict) -> None:
        """
        Check if the epics signal config is valid. The device must specify a read_pv.

        Args:
            name(str): name of the device
            config(dict): device config
        """
        self._has_auto_monitor(name, config["deviceConfig"])
        if "read_pv" not in config["deviceConfig"]:
            raise ValueError(f"{name}: does not specify the read_pv")

    def check_device_classes(self, name: str, conf: dict) -> int:
        """
        Run checks on the device class

        Args:
            name(str): name of the device
            conf(dict): device config

        Returns:
            int: 0 if all checks passed, 1 otherwise
        """
        try:
            dev_class = device_manager._get_device_class(conf["deviceClass"])

            if issubclass(dev_class, ophyd.EpicsMotor):
                self._check_epics_motor(name, conf)
                return 0

            if issubclass(dev_class, ophyd.signal.EpicsSignalBase):
                self._check_epics_signal(name, conf)
                return 0

            if issubclass(dev_class, ophyd.Device):
                self._check_all_signals_of_device(name, dev_class)
            return 0

        except Exception as e:
            self.print_and_write(f"ERROR: {name} is not valid: {e}")
            return 1

    def _has_auto_monitor(self, name: str, config: dict) -> None:
        """
        Check if the config has an auto_monitor key and print a warning if not.

        Args:
            name(str): name of the device
            config(dict): device config
        """
        if "auto_monitor" not in config:
            self.print_and_write(f"WARNING: Device {name} is configured without auto monitor.")

    def connect_device(self, name: str, conf: dict) -> int:
        """
        Connect to the device

        Args:
            name(str): name of the device
            conf(dict): device config

        Returns:
            int: 0 if all checks passed, 1 otherwise
        """
        try:
            conf_in = copy.deepcopy(conf)
            conf_in["name"] = name
            obj, _ = device_manager.construct_device_obj(conf_in, None)

            device_manager.connect_device(obj, wait_for_all=True)
            assert obj.connected is True
            self.check_basic_ophyd_methods(obj)
            obj.destroy()
            if hasattr(obj, "component_names") and obj.component_names:
                assert obj.connected is False
                assert obj._destroyed is True
            return 0

        except Exception:
            content = traceback.format_exc()
            self.print_and_write(f"ERROR: {name} is not connectable: {content}")
            return 1

    def check_basic_ophyd_methods(self, obj: ophyd.OphydObject) -> int:
        """
        Check if the basic ophyd methods work

        Args:
            obj(ophyd.OphydObject): device object

        Returns:

        """
        assert isinstance(obj, BECDevice)
        assert isinstance(obj.name, str)
        assert isinstance(obj.read(), dict)
        assert isinstance(obj.read_configuration(), dict)
        assert isinstance(obj.describe(), dict)
        assert isinstance(obj.describe_configuration(), dict)
        assert isinstance(obj.hints, dict)

    def validate_schema(self, name: str, conf: dict) -> None:
        """
        Validate the device config against the BEC DB schema

        Args:
            name(str): name of the device
            conf(dict): device config
        """
        try:
            validator = SciBecValidator()
            db_config = self._translate_to_db_config(name, conf)
            validator.validate_device(db_config)
            return 0
        except Exception as e:
            self.print_and_write(f"ERROR: {name} is not valid: {e}")
            return 1

    @staticmethod
    def _translate_to_db_config(name, config) -> dict:
        """
        Translate the device config to a db config

        Args:
            name(str): name of the device
            config(dict): device config

        Returns:
            dict: db config
        """
        db_config = copy.deepcopy(config)
        db_config["name"] = name
        if "deviceConfig" in db_config and db_config["deviceConfig"] is None:
            db_config["deviceConfig"] = {}
        db_config.pop("deviceType", None)
        return db_config

    def run(self, connect: bool) -> None:
        """
        Run the tests

        Args:
            connect(bool): connect to the devices
        """
        failed_devices = []
        for name, conf in self.config.items():
            return_val = 0
            self.print_and_write(f"Checking {name}...")
            return_val += self.validate_schema(name, conf)
            return_val += self.check_device_classes(name, conf)
            if connect:
                return_val += self.connect_device(name, conf)

            if return_val == 0:
                self.print_and_write("OK")
            else:
                self.print_and_write("FAILED")
                failed_devices.append(name)

        self.print_and_write("\n\n")
        self.print_and_write("========================================")
        # print summary
        self.print_and_write("Summary:")
        if len(failed_devices) == 0:
            print("All devices passed the test.")
            self.file.write("All devices passed the test.\n")
        else:
            print(f"{len(failed_devices)} devices failed the test:")
            self.file.write(f"{len(failed_devices)} devices failed the test:\n")
            for device in failed_devices:
                print(f"    {device}")
                self.file.write(f"    {device}\n")

    def print_and_write(self, text: str) -> None:
        """
        Print and write to the output file

        Args:
            text(str): text to print and write
        """
        print(text)
        self.file.write(text + "\n")


def launch() -> None:
    """launch the test"""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Perform tests on an ophyd device config file.",
    )

    parser.add_argument("--config", help="path to the config file", required=True, type=str)
    optional = parser.add_argument_group("optional arguments")
    optional.add_argument("--output", default="./report.txt", help="path to the output file")
    optional.add_argument("--connect", action="store_true", help="connect to the devices")
    parser.add_help = True

    clargs = parser.parse_args()

    if device_manager is None:
        raise ImportError(
            "device_server is not installed. Please install it first with pip install"
            " bec-device-server."
        )

    with open("./report.txt", "w", encoding="utf-8") as file:
        device_config_test = StaticDeviceTest(clargs.config, output_file=file)
        device_config_test.run(clargs.connect)


if __name__ == "__main__":  # pragma: no cover
    import sys

    sys.argv = ["", "--config", "../bec/bec_lib/bec_lib/configs/demo_config.yaml", "--connect"]
    launch()



from ophyd import Component, EpicsSignal, EpicsSignalRO, Kind, PVPositioner


class InsertionDevice(PVPositioner):
    """Python wrapper for the CSAXS insertion device control

    This wrapper provides a positioner interface for the ID control.
    is completely custom XBPM with templates directly in the
    VME repo. Thus it needs a custom ophyd template as well...

    WARN: The x and y are not updated by the IOC
    """

    status = Component(EpicsSignalRO, "-USER:STATUS", auto_monitor=True)
    errorSource = Component(EpicsSignalRO, "-USER:ERROR-SOURCE", auto_monitor=True)
    isOpen = Component(EpicsSignalRO, "-GAP:ISOPEN", auto_monitor=True)

    # PVPositioner interface
    setpoint = Component(EpicsSignal, "-GAP:SET", auto_monitor=True)
    readback = Component(EpicsSignalRO, "-GAP:READ", auto_monitor=True, kind=Kind.hinted)
    done = Component(EpicsSignalRO, ":DONE", auto_monitor=True)
    stop_signal = Component(EpicsSignal, "-GAP:STOP", kind=Kind.omitted)


# Automatically start simulation if directly invoked
# (NA for important devices)
if __name__ == "__main__":
    pass



from .sim import SimCamera
from .sim import SimFlyer
from .sim import SimFlyer as SynFlyer
from .sim import SimMonitor, SimPositioner, SimWaveform
from .sim_frameworks import SlitProxy
from .sim_signals import ReadOnlySignal, SetableSignal
from .sim_xtreme import SynXtremeOtf



# pylint: skip-file
from unittest import mock

import pytest

from ophyd_devices.epics.devices.delay_generator_csaxs import DDGSetup
from ophyd_devices.epics.devices.psi_delay_generator_base import TriggerSource


def patch_dual_pvs(device):
    for walk in device.walk_signals():
        if not hasattr(walk.item, "_read_pv"):
            continue
        if not hasattr(walk.item, "_write_pv"):
            continue
        if walk.item._read_pv.pvname.endswith("_RBV"):
            walk.item._read_pv = walk.item._write_pv


@pytest.fixture(scope="function")
def mock_DDGSetup():
    mock_ddg = mock.MagicMock()
    yield DDGSetup(parent=mock_ddg)


# Fixture for scaninfo
@pytest.fixture(
    params=[
        {
            "scan_id": "1234",
            "scan_type": "step",
            "num_points": 500,
            "frames_per_trigger": 1,
            "exp_time": 0.1,
            "readout_time": 0.1,
        },
        {
            "scan_id": "1234",
            "scan_type": "step",
            "num_points": 500,
            "frames_per_trigger": 5,
            "exp_time": 0.01,
            "readout_time": 0,
        },
        {
            "scan_id": "1234",
            "scan_type": "fly",
            "num_points": 500,
            "frames_per_trigger": 1,
            "exp_time": 1,
            "readout_time": 0.2,
        },
        {
            "scan_id": "1234",
            "scan_type": "fly",
            "num_points": 500,
            "frames_per_trigger": 5,
            "exp_time": 0.1,
            "readout_time": 0.4,
        },
    ]
)
def scaninfo(request):
    return request.param


# Fixture for DDG config default values
@pytest.fixture(
    params=[
        {
            "delay_burst": 0.0,
            "delta_width": 0.0,
            "additional_triggers": 0,
            "polarity": [0, 0, 0, 0, 0],
            "amplitude": 0.0,
            "offset": 0.0,
            "thres_trig_level": 0.0,
        },
        {
            "delay_burst": 0.1,
            "delta_width": 0.1,
            "additional_triggers": 1,
            "polarity": [0, 0, 1, 0, 0],
            "amplitude": 5,
            "offset": 0.0,
            "thres_trig_level": 2.5,
        },
    ]
)
def ddg_config_defaults(request):
    return request.param


# Fixture for DDG config scan values
@pytest.fixture(
    params=[
        {
            "fixed_ttl_width": [0, 0, 0, 0, 0],
            "trigger_width": None,
            "set_high_on_exposure": False,
            "set_high_on_stage": False,
            "set_trigger_source": "SINGLE_SHOT",
            "premove_trigger": False,
        },
        {
            "fixed_ttl_width": [0, 0, 0, 0, 0],
            "trigger_width": 0.1,
            "set_high_on_exposure": True,
            "set_high_on_stage": False,
            "set_trigger_source": "SINGLE_SHOT",
            "premove_trigger": True,
        },
        {
            "fixed_ttl_width": [0, 0, 0, 0, 0],
            "trigger_width": 0.1,
            "set_high_on_exposure": False,
            "set_high_on_stage": False,
            "set_trigger_source": "EXT_RISING_EDGE",
            "premove_trigger": False,
        },
    ]
)
def ddg_config_scan(request):
    return request.param


# Fixture for delay pairs
@pytest.fixture(
    params=[
        {"all_channels": ["channelAB", "channelCD"], "all_delay_pairs": ["AB", "CD"]},
        {"all_channels": [], "all_delay_pairs": []},
        {"all_channels": ["channelT0", "channelAB", "channelCD"], "all_delay_pairs": ["AB", "CD"]},
    ]
)
def channel_pairs(request):
    return request.param


def test_check_scan_id(mock_DDGSetup, scaninfo, ddg_config_defaults, ddg_config_scan):
    """Test the check_scan_id method."""
    # Set first attributes of parent class
    for k, v in scaninfo.items():
        setattr(mock_DDGSetup.parent.scaninfo, k, v)
    for k, v in ddg_config_defaults.items():
        getattr(mock_DDGSetup.parent, k).get.return_value = v
    for k, v in ddg_config_scan.items():
        getattr(mock_DDGSetup.parent, k).get.return_value = v
    # Call the function you want to test
    mock_DDGSetup.check_scan_id()
    mock_DDGSetup.parent.scaninfo.load_scan_metadata.assert_called_once()


def test_on_pre_scan(mock_DDGSetup, scaninfo, ddg_config_defaults, ddg_config_scan):
    """Test the check_scan_id method."""
    # Set first attributes of parent class
    for k, v in scaninfo.items():
        setattr(mock_DDGSetup.parent.scaninfo, k, v)
    for k, v in ddg_config_defaults.items():
        getattr(mock_DDGSetup.parent, k).get.return_value = v
    for k, v in ddg_config_scan.items():
        getattr(mock_DDGSetup.parent, k).get.return_value = v
    # Call the function you want to test
    mock_DDGSetup.on_pre_scan()
    if ddg_config_scan["premove_trigger"]:
        mock_DDGSetup.parent.trigger_shot.put.assert_called_once_with(1)


@pytest.mark.parametrize("source", ["SINGLE_SHOT", "EXT_RISING_EDGE"])
def test_on_trigger(mock_DDGSetup, scaninfo, ddg_config_defaults, ddg_config_scan, source):
    """Test the on_trigger method."""
    # Set first attributes of parent class
    for k, v in scaninfo.items():
        setattr(mock_DDGSetup.parent.scaninfo, k, v)
    for k, v in ddg_config_defaults.items():
        getattr(mock_DDGSetup.parent, k).get.return_value = v
    for k, v in ddg_config_scan.items():
        getattr(mock_DDGSetup.parent, k).get.return_value = v
    # Call the function you want to test
    mock_DDGSetup.parent.source.name = "source"
    mock_DDGSetup.parent.source.read.return_value = {
        mock_DDGSetup.parent.source.name: {"value": getattr(TriggerSource, source)}
    }
    mock_DDGSetup.on_trigger()
    if source == "SINGLE_SHOT":
        mock_DDGSetup.parent.trigger_shot.put.assert_called_once_with(1)


def test_initialize_default_parameter(
    mock_DDGSetup, scaninfo, ddg_config_defaults, ddg_config_scan, channel_pairs
):
    """Test the initialize_default_parameter method."""
    # Set first attributes of parent class
    for k, v in scaninfo.items():
        setattr(mock_DDGSetup.parent.scaninfo, k, v)
    for k, v in ddg_config_defaults.items():
        getattr(mock_DDGSetup.parent, k).get.return_value = v
    for k, v in ddg_config_scan.items():
        getattr(mock_DDGSetup.parent, k).get.return_value = v
    # Call the function you want to test
    mock_DDGSetup.parent.all_channels = channel_pairs["all_channels"]
    mock_DDGSetup.parent.all_delay_pairs = channel_pairs["all_delay_pairs"]
    calls = []
    calls.extend(
        [
            mock.call("polarity", ddg_config_defaults["polarity"][ii], [channel])
            for ii, channel in enumerate(channel_pairs["all_channels"])
        ]
    )
    calls.extend([mock.call("amplitude", ddg_config_defaults["amplitude"])])
    calls.extend([mock.call("offset", ddg_config_defaults["offset"])])
    calls.extend(
        [
            mock.call(
                "reference", 0, [f"channel{pair}.ch1" for pair in channel_pairs["all_delay_pairs"]]
            )
        ]
    )
    calls.extend(
        [
            mock.call(
                "reference", 0, [f"channel{pair}.ch2" for pair in channel_pairs["all_delay_pairs"]]
            )
        ]
    )
    mock_DDGSetup.initialize_default_parameter()
    mock_DDGSetup.parent.set_channels.assert_has_calls(calls)


def test_prepare_ddg(mock_DDGSetup, scaninfo, ddg_config_defaults, ddg_config_scan, channel_pairs):
    """Test the prepare_ddg method."""
    # Set first attributes of parent class
    for k, v in scaninfo.items():
        setattr(mock_DDGSetup.parent.scaninfo, k, v)
    for k, v in ddg_config_defaults.items():
        getattr(mock_DDGSetup.parent, k).get.return_value = v
    for k, v in ddg_config_scan.items():
        getattr(mock_DDGSetup.parent, k).get.return_value = v
    # Call the function you want to test
    mock_DDGSetup.parent.all_channels = channel_pairs["all_channels"]
    mock_DDGSetup.parent.all_delay_pairs = channel_pairs["all_delay_pairs"]

    mock_DDGSetup.prepare_ddg()
    mock_DDGSetup.parent.set_trigger.assert_called_once_with(
        getattr(TriggerSource, ddg_config_scan["set_trigger_source"])
    )
    if scaninfo["scan_type"] == "step":
        if ddg_config_scan["set_high_on_exposure"]:
            num_burst_cycle = 1 + ddg_config_defaults["additional_triggers"]
            exp_time = ddg_config_defaults["delta_width"] + scaninfo["frames_per_trigger"] * (
                scaninfo["exp_time"] + scaninfo["readout_time"]
            )
            total_exposure = exp_time
            delay_burst = ddg_config_defaults["delay_burst"]
        else:
            exp_time = ddg_config_defaults["delta_width"] + scaninfo["exp_time"]
            total_exposure = exp_time + scaninfo["readout_time"]
            delay_burst = ddg_config_defaults["delay_burst"]
            num_burst_cycle = (
                scaninfo["frames_per_trigger"] + ddg_config_defaults["additional_triggers"]
            )
    elif scaninfo["scan_type"] == "fly":
        if ddg_config_scan["set_high_on_exposure"]:
            num_burst_cycle = 1 + ddg_config_defaults["additional_triggers"]
            exp_time = (
                ddg_config_defaults["delta_width"]
                + scaninfo["num_points"] * scaninfo["exp_time"]
                + (scaninfo["num_points"] - 1) * scaninfo["readout_time"]
            )
            total_exposure = exp_time
            delay_burst = ddg_config_defaults["delay_burst"]
        else:
            exp_time = ddg_config_defaults["delta_width"] + scaninfo["exp_time"]
            total_exposure = exp_time + scaninfo["readout_time"]
            delay_burst = ddg_config_defaults["delay_burst"]
            num_burst_cycle = scaninfo["num_points"] + ddg_config_defaults["additional_triggers"]

    # mock_DDGSetup.parent.burst_enable.assert_called_once_with(
    #     mock.call(num_burst_cycle, delay_burst, total_exposure, config="first")
    # )
    mock_DDGSetup.parent.burst_enable.assert_called_once_with(
        num_burst_cycle, delay_burst, total_exposure, config="first"
    )
    if not ddg_config_scan["trigger_width"]:
        call = mock.call("width", exp_time)
        assert call in mock_DDGSetup.parent.set_channels.mock_calls
    else:
        call = mock.call("width", ddg_config_scan["trigger_width"])
        assert call in mock_DDGSetup.parent.set_channels.mock_calls
    if ddg_config_scan["set_high_on_exposure"]:
        calls = [
            mock.call("width", value, channels=[channel])
            for value, channel in zip(
                ddg_config_scan["fixed_ttl_width"], channel_pairs["all_channels"]
            )
            if value != 0
        ]
        if calls:
            assert all(calls in mock_DDGSetup.parent.set_channels.mock_calls)



from enum import Enum


class TomcatSequencerState:
    IDLE = 0
    READY = 1
    ARMED = 2
    RUNNING = 3
    CHANGING = 4
    ERROR = 15


class AxisDataSignal:
    PositionFeedback = 0
    PositionCommand = 1
    PositionError = 2
    VelocityFeedback = 3
    VelocityCommand = 4
    VelocityError = 5
    AccelerationCommand = 6
    CurrentError = 9
    PositionCommandRaw = 12
    VelocityCommandRaw = 13
    AuxiliaryFeedback = 14
    DigitalInput = 15
    DigitalOutput = 16
    FixtureOffset = 18
    CoordinatedPositionTarget = 41
    DriveStatus = 42
    AxisStatus = 43
    AxisFault = 44
    AccelerationCommandRaw = 45
    PositionCalibrationAll = 50
    PositionFeedbackRollover = 63
    PositionCommandRollover = 64
    VelocityFeedbackAverage = 65
    CurrentFeedbackAverage = 66
    AxisParameter = 68
    Backlash = 72
    HomeState = 73
    PositionCalibration2D = 74
    NormalcyDebug = 75
    TotalMoveTime = 76
    JerkCommandRaw = 78
    ProgramPositionCommand = 79
    PositionOffset = 80
    PositionCommandRawBackwardsDiff = 82
    VelocityCommandRawBackwardsDiffDelta = 83
    DriveStatusActual = 85
    ProgramPositionFeedback = 89
    JogTrajectoryStatus = 94
    PingTest = 95
    AccelerationTime = 109
    DecelerationTime = 110
    AccelerationRate = 111
    DecelerationRate = 112
    AccelerationType = 113
    DecelerationType = 114
    AccelerationMode = 115
    DecelerationMode = 116
    ProgramPosition = 124
    SpeedTarget = 128
    PositionCommandPacket = 131
    DriveSmcMotionState = 132
    PositionCommandRawCal = 140
    VelocityCommandRawCal = 141
    VelocityCommandDrive = 142
    AccelerationCommandDrive = 143
    GalvoLaserOutputRaw = 144
    DriveInterfacePacketInt32 = 147
    DriveInterfacePacketInt16 = 148
    DriveInterfacePacketInt8 = 149
    DriveInterfacePacketDouble = 150
    DriveInterfacePacketFloat = 151
    DriveInterfaceCommandCode = 152
    AccelerationFeedback = 153
    AccelerationCommandRawCal = 154
    PositionCalibrationAllDrive = 155
    BacklashTarget = 156
    DriveMotionRate = 158
    DriveMotionDelay = 159
    CalibrationAdjustmentValue = 160
    ServoRounding = 161
    FeedforwardCurrent = 162
    DriveInterfacePacketInfoBitValue = 164
    AccelerationError = 165
    SuppressedFaults = 167
    DriveInterfacePacketStreamingData = 168
    PositionCommandRawUnfiltered = 169
    TransitionOffsetErrors = 170
    FreezeVelocityCommand = 179
    FreezeVelocityFeedback = 180
    InternalPositionOffsets = 181
    StatusHighLevelOffsetsLastMsec = 182
    ProgramVelocityCommand = 183
    ProgramVelocityFeedback = 184
    DriveMotionDelayLive = 185
    DriveCommunicationDelay = 186
    DriveCommunicationDelayLive = 187
    DriveInterfacePacketResponseInt32 = 189
    DriveInterfacePacketResponseInt16 = 190
    DriveInterfacePacketResponseInt8 = 191
    DriveInterfacePacketResponseDouble = 192
    DriveInterfacePacketResponseFloat = 193
    DriveInterfacePacketBit = 194
    DriveInterfacePacketResponseBit = 195
    SpeedTargetActual = 196
    CoordinatedDistanceRemaining = 199
    SafeZoneState = 230
    PositionErrorGalvo = 235
    MoveReferencePosition = 237
    MoveReferenceCutterOffset = 250
    MoveReferenceCornerOffset = 251
    MoveReferenceTotalOffset = 252
    DistanceLog = 264
    AutoFocusError = 295
    GalvoLaserOutputRawAdvance = 296
    GalvoLaserOnDelay = 297
    GalvoLaserOffDelay = 298
    CalibrationAdjustmentState = 301
    AccuracyCorrectionStartingPosition = 302
    AccuracyCorrectionEndingPosition = 303
    DriveCommandsDelayed = 309
    DriveCommandsLost = 310
    StoStatus = 327
    DriveAssert = 336
    PrimaryFeedback = 366
    AccelerationSCurvePercentage = 371
    DecelerationSCurvePercentage = 372
    GantryMarkerDifference = 390
    PrimaryFeedbackStatus = 392
    HomeTargetPosition = 398
    GantryRealignmentMoveTargetPosition = 399
    GantryDriveControlRealignmentState = 400
    DriveInterfacePositionCommandPhysical = 434
    DriveControlReason = 435
    CurrentFeedback = 7
    CurrentCommand = 8
    AnalogInput0 = 10
    AnalogInput1 = 11
    PhaseACurrentFeedback = 19
    PhaseBCurrentFeedback = 20
    EncoderSine = 21
    EncoderCosine = 22
    AnalogInput2 = 23
    AnalogInput3 = 24
    FrequencyResponseBefore = 25
    FrequencyResponseAfter = 26
    AnalogOutput0 = 31
    AnalogOutput1 = 32
    AnalogOutput2 = 33
    AnalogOutput3 = 34
    DriveMemoryInt32 = 35
    DriveMemoryFloat = 36
    DriveMemoryDouble = 37
    PsoStatus = 38
    DriveTimerDebug = 39
    PositionFeedbackDrive = 77
    PositionCommandDrive = 84
    DriveMemoryInt16 = 125
    DriveMemoryInt8 = 126
    PsoCounter0 = 171
    PsoCounter1 = 172
    PsoCounter2 = 173
    PsoWindow0 = 174
    PsoWindow1 = 175
    DriveDataCaptureSamples = 176
    PositionCommandGalvo = 178
    PrimaryEnDatAbsolutePosition = 197
    ControlEffort = 201
    PhaseAVoltageCommand = 208
    PhaseBVoltageCommand = 209
    PhaseCVoltageCommand = 210
    AmplifierPeakCurrent = 211
    FpgaVersion = 212
    DriveTypeId = 213
    PsoWindow0ArrayIndex = 214
    PsoWindow1ArrayIndex = 215
    PsoDistanceArrayIndex = 216
    AmplifierTemperature = 217
    PsoBitArrayIndex = 218
    MxAbsolutePosition = 219
    ServoUpdateRate = 220
    SettlingTime = 221
    InternalStatusCode = 222
    FirmwareVersionMajor = 223
    FirmwareVersionMinor = 224
    FirmwareVersionPatch = 225
    FirmwareVersionBuild = 226
    DriveTimerDebugMax = 227
    MarkerSearchDistance = 228
    PositionFeedbackGalvo = 234
    LatchedMarkerPosition = 236
    PrimaryBissAbsolutePosition = 255
    FaultPositionFeedback = 258
    MotorCommutationAngle = 259
    ExpansionBoardOption = 260
    BusVoltage = 261
    PiezoVoltageCommand = 262
    PiezoVoltageFeedback = 263
    TimeSinceReset = 273
    MaximumVoltage = 274
    CommandOutputType = 275
    DriveFeedforwardOutput = 290
    LastTickCounter = 291
    BoardRevision = 292
    GalvoLaserOutput = 294
    GalvoLaserPowerCorrectionOutput = 299
    CapacitanceSensorRawPosition = 300
    PositionCalibrationGalvo = 304
    BusVoltageNegative = 325
    ProcessorTemperature = 326
    InternalStatusTimestamp = 328
    AnalogSensorInput = 329
    MotorTemperature = 330
    PrimaryBissStatus = 332
    PsoExternalSyncFrequency = 337
    EncoderSineRaw = 346
    EncoderCosineRaw = 347
    FpgaTemperature = 353
    PrimaryEnDatStatus = 355
    DriveTimerHighPriorityThread = 356
    DriveTimerLowPriorityThread = 357
    DriveTimerLowPriorityPacket = 358
    DriveTimerServoPacket = 359
    DriveTimerServoThread = 360
    DriveTimerCurrentPacket = 361
    DriveTimerCommonCoreThread = 362
    DriveTimerServoCorePacket0 = 363
    DriveTimerServoCorePacket1 = 364
    MultiplierOption = 365
    ServoLoopFeedbackInput0 = 367
    ServoLoopFeedbackInput1 = 368
    FaultSubcode = 376
    ProcessorTemperatureMax = 378
    DriveTimerHyperWireDma = 381
    AmplifierTemperatureMax = 382
    AuxiliaryEnDatAbsolutePosition = 383
    AuxiliaryEnDatStatus = 384
    AuxiliaryBissAbsolutePosition = 385
    AuxiliaryBissStatus = 386
    PsoOption = 387
    DriveArraySize = 388
    RatedMotorSupplyVoltageOption = 389
    AbsoluteEncoderOption = 391
    AuxiliaryFeedbackStatus = 393
    AmplifierStatus = 394
    LatchedCwLimitPosition = 395
    LatchedCcwLimitPosition = 396
    GalvoLaserFpgaTransitionDelay = 397
    PiezoAccumulatedCharge = 401
    PiezoChargingTime = 402
    PrimarySsiAbsolutePosition = 403
    PrimarySsiStatus = 404
    AuxiliarySsiAbsolutePosition = 405
    AuxiliarySsiStatus = 406
    PsoDistanceActiveDistance = 407
    PsoWindow0ActiveLowerBound = 408
    PsoWindow0ActiveUpperBound = 409
    PsoWindow1ActiveLowerBound = 410
    PsoWindow1ActiveUpperBound = 411
    PsoWaveformActiveTotalTime = 412
    PsoWaveformActiveOnTime = 413
    PsoWaveformActivePulseCount = 414
    PsoEventActiveBitValue = 415
    DriveTimerDriveBasedControllerOutputDma = 419
    DriveTimerPcieInboundFsm = 420
    PrimaryFeedbackServo = 425
    AuxiliaryFeedbackServo = 426
    DriveStackUsage = 427
    ShuntResistorTemperature = 436


class TaskDataSignal:
    ProgramLineNumber = 17
    CoordinatedFlags = 40
    CoordinatedArcStartAngle = 53
    CoordinatedArcEndAngle = 54
    CoordinatedArcRadius = 55
    CoordinatedArcRadiusError = 56
    CoordinatedPositionCommand = 57
    CoordinatedSpeedCommand = 58
    CoordinatedAccelerationCommand = 59
    CoordinatedTotalDistance = 60
    CoordinatedPercentDone = 61
    CoordinatedPositionCommandBackwardsDiff = 62
    TaskParameter = 69
    TaskError = 70
    TaskWarning = 71
    CoordinatedSpeedTargetActual = 86
    DependentCoordinatedSpeedTargetActual = 87
    ActiveFixtureOffset = 88
    TaskStatus0 = 90
    TaskStatus1 = 91
    TaskStatus2 = 92
    SpindleSpeedTarget = 93
    CoordinateSystem1I = 96
    CoordinateSystem1J = 97
    CoordinateSystem1K = 98
    CoordinateSystem1Plane = 99
    ToolNumberActive = 100
    Mfo = 101
    CoordinatedSpeedTarget = 102
    DependentCoordinatedSpeedTarget = 103
    CoordinatedAccelerationRate = 104
    CoordinatedDecelerationRate = 105
    CoordinatedAccelerationTime = 106
    CoordinatedDecelerationTime = 107
    TaskMode = 108
    TaskState = 117
    TaskStateInternal = 118
    ExecutionMode = 121
    EnableAlignmentAxes = 127
    CoordinatedGalvoLaserOutput = 133
    CoordinatedMotionRate = 145
    CoordinatedTaskCommand = 146
    EnableState = 166
    LookaheadMovesExamined = 200
    ProfileControlMask = 231
    CoordinatedArcRadiusReciprocal = 253
    MotionEngineStage = 254
    CoordinatedTimeScale = 256
    CoordinatedTimeScaleDerivative = 257
    IfovSpeedScale = 266
    IfovSpeedScaleAverage = 267
    IfovGenerationFrameCounter = 268
    IfovGenerationTimeOriginal = 269
    IfovGenerationTimeModified = 270
    IfovCoordinatedPositionCommand = 271
    IfovCoordinatedSpeedCommand = 272
    IfovCenterPointH = 276
    IfovCenterPointV = 277
    IfovTrajectoryCount = 278
    IfovTrajectoryIndex = 279
    IfovAttemptCode = 280
    IfovGenerationFrameIndex = 281
    IfovMaximumVelocity = 282
    IfovIdealVelocity = 283
    TaskInternalDebug = 284
    IfovCoordinatedAccelerationCommand = 285
    IfovFovPositionH = 286
    IfovFovPositionV = 287
    IfovFovDimensionH = 288
    IfovFovDimensionV = 289
    MotionBufferElements = 311
    MotionBufferMoves = 312
    MotionLineNumber = 313
    MotionBufferRetraceMoves = 314
    MotionBufferRetraceElements = 315
    MotionBufferIndex = 316
    MotionBufferIndexLookahead = 317
    MotionBufferProcessingBlocked = 318
    ActiveMoveValid = 319
    TaskExecutionLines = 320
    SchedulerTaskHolds = 321
    SchedulerProgramLoopRuns = 322
    SchedulerTaskBlocked = 323
    CriticalSectionsActive = 324
    AxesSlowdownReason = 331
    TaskExecutionTime = 333
    TaskExecutionTimeMaximum = 334
    TaskExecutionLinesMaximum = 335
    LookaheadDecelReason = 338
    LookaheadDecelMoves = 339
    LookaheadDecelDistance = 340
    ProgramCounter = 341
    StackPointer = 342
    FramePointer = 343
    StringStackPointer = 344
    ProgramLineNumberSourceFileId = 349
    MotionLineNumberSourceFileId = 350
    ProgramLineNumberSourcePathId = 351
    MotionLineNumberSourcePathId = 352
    StringArgumentStackPointer = 354
    CoordinatedAccelerationSCurvePercentage = 369
    CoordinatedDecelerationSCurvePercentage = 370
    DependentCoordinatedAccelerationRate = 373
    DependentCoordinatedDecelerationRate = 374
    CriticalSectionTimeout = 375
    CommandQueueCapacity = 421
    CommandQueueUnexecutedCount = 422
    CommandQueueTimesEmptied = 423
    CommandQueueExecutedCount = 424


class SystemDataSignal:
    VirtualBinaryInput = 46
    VirtualBinaryOutput = 47
    VirtualRegisterInput = 48
    VirtualRegisterOutput = 49
    Timer = 51
    TimerPerformance = 52
    GlobalReal = 67
    CommunicationRealTimeErrors = 81
    LibraryCommand = 119
    DataCollectionSampleTime = 120
    DataCollectionSampleIndex = 129
    ModbusClientConnected = 134
    ModbusServerConnected = 135
    ModbusClientError = 136
    ModbusServerError = 137
    StopWatchTimer = 157
    ScopetrigId = 163
    EstimatedProcessorUsage = 177
    DataCollectionStatus = 188
    SignalLogState = 198
    SafeZoneViolationMask = 207
    SafeZoneActiveMask = 229
    ModbusClientInputWords = 240
    ModbusClientOutputWords = 241
    ModbusClientInputBits = 242
    ModbusClientOutputBits = 243
    ModbusClientOutputBitsStatus = 244
    ModbusClientOutputWordsStatus = 245
    ModbusServerInputWords = 246
    ModbusServerOutputWords = 247
    ModbusServerInputBits = 248
    ModbusServerOutputBits = 249
    SystemParameter = 265
    ThermoCompSensorTemperature = 305
    ThermoCompControllingTemperature = 306
    ThermoCompCompensatingTemperature = 307
    ThermoCompStatus = 308
    GlobalInteger = 345
    AliveAxesMask = 348
    SignalLogPointsStored = 377
    ControllerInitializationWarning = 379
    StopWatchTimerMin = 416
    StopWatchTimerMax = 417
    StopWatchTimerAvg = 418
    EthercatEnabled = 428
    EthercatError = 429
    EthercatTxPdo = 430
    EthercatTxPdoSize = 431
    EthercatRxPdo = 432
    EthercatRxPdoSize = 433
    EthercatState = 437
    ModbusClientEnabled = 438
    ModbusServerEnabled = 439


class DataCollectionFrequency:
    Undefined = 0
    Fixed1kHz = 1
    Fixed10kHz = 2
    Fixed20kHz = 3
    Fixed100kHz = 4
    Fixed200kHz = 5


class DataCollectionMode:
    Snapshot = 0
    Continouous = 1


# Specifies the PSO distance input settings for the XC4e drive.
class PsoDistanceInput:
    XC4PrimaryFeedback = 130
    XC4AuxiliaryFeedback = 131
    XC4SyncPortA = 132
    XC4SyncPortB = 133
    XC4DrivePulseStream = 134
    XC4ePrimaryFeedback = 135
    XC4eAuxiliaryFeedback = 136
    XC4eSyncPortA = 137
    XC4eSyncPortB = 138
    XC4eDrivePulseStream = 139


class PsoWindowInput:
    XC4PrimaryFeedback = 130
    XC4AuxiliaryFeedback = 131
    XC4SyncPortA = 132
    XC4SyncPortB = 133
    XC4DrivePulseStream = 134
    XC4ePrimaryFeedback = 135
    XC4eAuxiliaryFeedback = 136
    XC4eSyncPortA = 137
    XC4eSyncPortB = 138
    XC4eDrivePulseStream = 139
    XL5ePrimaryFeedback = (145,)
    XL5eAuxiliaryFeedback = (146,)
    XL5eSyncPortA = (147,)
    XL5eSyncPortB = (148,)
    XL5eDrivePulseStream = (149,)


# @brief Specifies the PSO output pin settings for each drive.
class XC4ePsoOutputPin:
    DedicatedOutput = 111
    AuxiliaryMarkerDifferential = 112
    AuxiliaryMarkerSingleEnded = 113


class XC4PsoOutputPin:
    DedicatedOutput = 108
    AuxiliaryMarkerDifferential = 109
    AuxiliaryMarkerSingleEnded = 110


"""
# @brief Specifies the PSO distance input settings for each drive.
class Automation1PsoDistanceInput:
	Automation1PsoDistanceInput_GL4PrimaryFeedbackAxis1 = 100,
	Automation1PsoDistanceInput_GL4PrimaryFeedbackAxis2 = 101,
	Automation1PsoDistanceInput_GL4IfovFeedbackAxis1 = 102,
	Automation1PsoDistanceInput_GL4IfovFeedbackAxis2 = 103,
	Automation1PsoDistanceInput_GL4AuxiliaryFeedbackAxis1 = 104,
	Automation1PsoDistanceInput_GL4AuxiliaryFeedbackAxis2 = 105,
	Automation1PsoDistanceInput_GL4SyncPortA = 106,
	Automation1PsoDistanceInput_GL4SyncPortB = 107,
	Automation1PsoDistanceInput_GL4DrivePulseStreamAxis1 = 108,
	Automation1PsoDistanceInput_GL4DrivePulseStreamAxis2 = 109,
	Automation1PsoDistanceInput_XL4sPrimaryFeedback = 110,
	Automation1PsoDistanceInput_XL4sAuxiliaryFeedback = 111,
	Automation1PsoDistanceInput_XL4sSyncPortA = 112,
	Automation1PsoDistanceInput_XL4sSyncPortB = 113,
	Automation1PsoDistanceInput_XL4sDrivePulseStream = 114,
	Automation1PsoDistanceInput_XR3PrimaryFeedbackAxis1 = 115,
	Automation1PsoDistanceInput_XR3PrimaryFeedbackAxis2 = 116,
	Automation1PsoDistanceInput_XR3PrimaryFeedbackAxis3 = 117,
	Automation1PsoDistanceInput_XR3PrimaryFeedbackAxis4 = 118,
	Automation1PsoDistanceInput_XR3PrimaryFeedbackAxis5 = 119,
	Automation1PsoDistanceInput_XR3PrimaryFeedbackAxis6 = 120,
	Automation1PsoDistanceInput_XR3AuxiliaryFeedbackAxis1 = 121,
	Automation1PsoDistanceInput_XR3AuxiliaryFeedbackAxis2 = 122,
	Automation1PsoDistanceInput_XR3AuxiliaryFeedbackAxis3 = 123,
	Automation1PsoDistanceInput_XR3AuxiliaryFeedbackAxis4 = 124,
	Automation1PsoDistanceInput_XR3AuxiliaryFeedbackAxis5 = 125,
	Automation1PsoDistanceInput_XR3AuxiliaryFeedbackAxis6 = 126,
	Automation1PsoDistanceInput_XR3SyncPortA = 127,
	Automation1PsoDistanceInput_XR3SyncPortB = 128,
	Automation1PsoDistanceInput_XR3DrivePulseStream = 129,
	Automation1PsoDistanceInput_XC4PrimaryFeedback = 130,
	Automation1PsoDistanceInput_XC4AuxiliaryFeedback = 131,
	Automation1PsoDistanceInput_XC4SyncPortA = 132,
	Automation1PsoDistanceInput_XC4SyncPortB = 133,
	Automation1PsoDistanceInput_XC4DrivePulseStream = 134,
	XC4ePrimaryFeedback = 135,
	XC4eAuxiliaryFeedback = 136,
	XC4eSyncPortA = 137,
	XC4eSyncPortB = 138,
	XC4eDrivePulseStream = 139,
	Automation1PsoDistanceInput_XC6ePrimaryFeedback = 140,
	Automation1PsoDistanceInput_XC6eAuxiliaryFeedback = 141,
	Automation1PsoDistanceInput_XC6eSyncPortA = 142,
	Automation1PsoDistanceInput_XC6eSyncPortB = 143,
	Automation1PsoDistanceInput_XC6eDrivePulseStream = 144,
	Automation1PsoDistanceInput_XL5ePrimaryFeedback = 145,
	Automation1PsoDistanceInput_XL5eAuxiliaryFeedback = 146,
	Automation1PsoDistanceInput_XL5eSyncPortA = 147,
	Automation1PsoDistanceInput_XL5eSyncPortB = 148,
	Automation1PsoDistanceInput_XL5eDrivePulseStream = 149,
	Automation1PsoDistanceInput_XC2PrimaryFeedback = 150,
	Automation1PsoDistanceInput_XC2AuxiliaryFeedback = 151,
	Automation1PsoDistanceInput_XC2DrivePulseStream = 152,
	Automation1PsoDistanceInput_XC2ePrimaryFeedback = 153,
	Automation1PsoDistanceInput_XC2eAuxiliaryFeedback = 154,
	Automation1PsoDistanceInput_XC2eDrivePulseStream = 155,
	Automation1PsoDistanceInput_XL2ePrimaryFeedback = 156,
	Automation1PsoDistanceInput_XL2eAuxiliaryFeedback = 157,
	Automation1PsoDistanceInput_XL2eSyncPortA = 158,
	Automation1PsoDistanceInput_XL2eSyncPortB = 159,
	Automation1PsoDistanceInput_XL2eDrivePulseStream = 160,
	Automation1PsoDistanceInput_XI4PrimaryFeedbackAxis1 = 161,
	Automation1PsoDistanceInput_XI4PrimaryFeedbackAxis2 = 162,
	Automation1PsoDistanceInput_XI4PrimaryFeedbackAxis3 = 163,
	Automation1PsoDistanceInput_XI4PrimaryFeedbackAxis4 = 164,
	Automation1PsoDistanceInput_XI4AuxiliaryFeedback1 = 165,
	Automation1PsoDistanceInput_XI4AuxiliaryFeedback2 = 166,
	Automation1PsoDistanceInput_XI4AuxiliaryFeedback3 = 167,
	Automation1PsoDistanceInput_XI4AuxiliaryFeedback4 = 168,
	Automation1PsoDistanceInput_XI4SyncPortA = 169,
	Automation1PsoDistanceInput_XI4SyncPortB = 170,
	Automation1PsoDistanceInput_XI4DrivePulseStreamAxis1 = 171,
	Automation1PsoDistanceInput_XI4DrivePulseStreamAxis2 = 172,
	Automation1PsoDistanceInput_XI4DrivePulseStreamAxis3 = 173,
	Automation1PsoDistanceInput_XI4DrivePulseStreamAxis4 = 174,
	Automation1PsoDistanceInput_iXC4PrimaryFeedback = 175,
	Automation1PsoDistanceInput_iXC4AuxiliaryFeedback = 176,
	Automation1PsoDistanceInput_iXC4SyncPortA = 177,
	Automation1PsoDistanceInput_iXC4SyncPortB = 178,
	Automation1PsoDistanceInput_iXC4DrivePulseStream = 179,
	Automation1PsoDistanceInput_iXC4ePrimaryFeedback = 180,
	Automation1PsoDistanceInput_iXC4eAuxiliaryFeedback = 181,
	Automation1PsoDistanceInput_iXC4eSyncPortA = 182,
	Automation1PsoDistanceInput_iXC4eSyncPortB = 183,
	Automation1PsoDistanceInput_iXC4eDrivePulseStream = 184,
	Automation1PsoDistanceInput_iXC6ePrimaryFeedback = 185,
	Automation1PsoDistanceInput_iXC6eAuxiliaryFeedback = 186,
	Automation1PsoDistanceInput_iXC6eSyncPortA = 187,
	Automation1PsoDistanceInput_iXC6eSyncPortB = 188,
	Automation1PsoDistanceInput_iXC6eDrivePulseStream = 189,
	Automation1PsoDistanceInput_iXL5ePrimaryFeedback = 190,
	Automation1PsoDistanceInput_iXL5eAuxiliaryFeedback = 191,
	Automation1PsoDistanceInput_iXL5eSyncPortA = 192,
	Automation1PsoDistanceInput_iXL5eSyncPortB = 193,
	Automation1PsoDistanceInput_iXL5eDrivePulseStream = 194,
	Automation1PsoDistanceInput_iXR3PrimaryFeedbackAxis1 = 195,
	Automation1PsoDistanceInput_iXR3PrimaryFeedbackAxis2 = 196,
	Automation1PsoDistanceInput_iXR3PrimaryFeedbackAxis3 = 197,
	Automation1PsoDistanceInput_iXR3PrimaryFeedbackAxis4 = 198,
	Automation1PsoDistanceInput_iXR3PrimaryFeedbackAxis5 = 199,
	Automation1PsoDistanceInput_iXR3PrimaryFeedbackAxis6 = 200,
	Automation1PsoDistanceInput_iXR3AuxiliaryFeedbackAxis1 = 201,
	Automation1PsoDistanceInput_iXR3AuxiliaryFeedbackAxis2 = 202,
	Automation1PsoDistanceInput_iXR3AuxiliaryFeedbackAxis3 = 203,
	Automation1PsoDistanceInput_iXR3AuxiliaryFeedbackAxis4 = 204,
	Automation1PsoDistanceInput_iXR3AuxiliaryFeedbackAxis5 = 205,
	Automation1PsoDistanceInput_iXR3AuxiliaryFeedbackAxis6 = 206,
	Automation1PsoDistanceInput_iXR3SyncPortA = 207,
	Automation1PsoDistanceInput_iXR3SyncPortB = 208,
	Automation1PsoDistanceInput_iXR3DrivePulseStream = 209,
	Automation1PsoDistanceInput_GI4DrivePulseStreamAxis1 = 210,
	Automation1PsoDistanceInput_GI4DrivePulseStreamAxis2 = 211,
	Automation1PsoDistanceInput_GI4DrivePulseStreamAxis3 = 212,
	Automation1PsoDistanceInput_iXC2PrimaryFeedback = 213,
	Automation1PsoDistanceInput_iXC2AuxiliaryFeedback = 214,
	Automation1PsoDistanceInput_iXC2DrivePulseStream = 215,
	Automation1PsoDistanceInput_iXC2ePrimaryFeedback = 216,
	Automation1PsoDistanceInput_iXC2eAuxiliaryFeedback = 217,
	Automation1PsoDistanceInput_iXC2eDrivePulseStream = 218,
	Automation1PsoDistanceInput_iXL2ePrimaryFeedback = 219,
	Automation1PsoDistanceInput_iXL2eAuxiliaryFeedback = 220,
	Automation1PsoDistanceInput_iXL2eSyncPortA = 221,
	Automation1PsoDistanceInput_iXL2eSyncPortB = 222,
	Automation1PsoDistanceInput_iXL2eDrivePulseStream = 223,
	Automation1PsoDistanceInput_iXI4PrimaryFeedbackAxis1 = 224,
	Automation1PsoDistanceInput_iXI4PrimaryFeedbackAxis2 = 225,
	Automation1PsoDistanceInput_iXI4PrimaryFeedbackAxis3 = 226,
	Automation1PsoDistanceInput_iXI4PrimaryFeedbackAxis4 = 227,
	Automation1PsoDistanceInput_iXI4AuxiliaryFeedback1 = 228,
	Automation1PsoDistanceInput_iXI4AuxiliaryFeedback2 = 229,
	Automation1PsoDistanceInput_iXI4AuxiliaryFeedback3 = 230,
	Automation1PsoDistanceInput_iXI4AuxiliaryFeedback4 = 231,
	Automation1PsoDistanceInput_iXI4SyncPortA = 232,
	Automation1PsoDistanceInput_iXI4SyncPortB = 233,
	Automation1PsoDistanceInput_iXI4DrivePulseStreamAxis1 = 234,
	Automation1PsoDistanceInput_iXI4DrivePulseStreamAxis2 = 235,
	Automation1PsoDistanceInput_iXI4DrivePulseStreamAxis3 = 236,
	Automation1PsoDistanceInput_iXI4DrivePulseStreamAxis4 = 237,

# @brief Specifies the PSO window input settings for each drive.
class Automation1PsoWindowInput:
	Automation1PsoWindowInput_GL4PrimaryFeedbackAxis1 = 100,
	Automation1PsoWindowInput_GL4PrimaryFeedbackAxis2 = 101,
	Automation1PsoWindowInput_GL4IfovFeedbackAxis1 = 102,
	Automation1PsoWindowInput_GL4IfovFeedbackAxis2 = 103,
	Automation1PsoWindowInput_GL4AuxiliaryFeedbackAxis1 = 104,
	Automation1PsoWindowInput_GL4AuxiliaryFeedbackAxis2 = 105,
	Automation1PsoWindowInput_GL4SyncPortA = 106,
	Automation1PsoWindowInput_GL4SyncPortB = 107,
	Automation1PsoWindowInput_GL4DrivePulseStreamAxis1 = 108,
	Automation1PsoWindowInput_GL4DrivePulseStreamAxis2 = 109,
	Automation1PsoWindowInput_XL4sPrimaryFeedback = 110,
	Automation1PsoWindowInput_XL4sAuxiliaryFeedback = 111,
	Automation1PsoWindowInput_XL4sSyncPortA = 112,
	Automation1PsoWindowInput_XL4sSyncPortB = 113,
	Automation1PsoWindowInput_XL4sDrivePulseStream = 114,
	Automation1PsoWindowInput_XR3PrimaryFeedbackAxis1 = 115,
	Automation1PsoWindowInput_XR3PrimaryFeedbackAxis2 = 116,
	Automation1PsoWindowInput_XR3PrimaryFeedbackAxis3 = 117,
	Automation1PsoWindowInput_XR3PrimaryFeedbackAxis4 = 118,
	Automation1PsoWindowInput_XR3PrimaryFeedbackAxis5 = 119,
	Automation1PsoWindowInput_XR3PrimaryFeedbackAxis6 = 120,
	Automation1PsoWindowInput_XR3AuxiliaryFeedbackAxis1 = 121,
	Automation1PsoWindowInput_XR3AuxiliaryFeedbackAxis2 = 122,
	Automation1PsoWindowInput_XR3AuxiliaryFeedbackAxis3 = 123,
	Automation1PsoWindowInput_XR3AuxiliaryFeedbackAxis4 = 124,
	Automation1PsoWindowInput_XR3AuxiliaryFeedbackAxis5 = 125,
	Automation1PsoWindowInput_XR3AuxiliaryFeedbackAxis6 = 126,
	Automation1PsoWindowInput_XR3SyncPortA = 127,
	Automation1PsoWindowInput_XR3SyncPortB = 128,
	Automation1PsoWindowInput_XR3DrivePulseStream = 129,
	Automation1PsoWindowInput_XC4PrimaryFeedback = 130,
	Automation1PsoWindowInput_XC4AuxiliaryFeedback = 131,
	Automation1PsoWindowInput_XC4SyncPortA = 132,
	Automation1PsoWindowInput_XC4SyncPortB = 133,
	Automation1PsoWindowInput_XC4DrivePulseStream = 134,
	XC4ePrimaryFeedback = 135,
	XC4eAuxiliaryFeedback = 136,
	XC4eSyncPortA = 137,
	XC4eSyncPortB = 138,
	XC4eDrivePulseStream = 139,
	Automation1PsoWindowInput_XC6ePrimaryFeedback = 140,
	Automation1PsoWindowInput_XC6eAuxiliaryFeedback = 141,
	Automation1PsoWindowInput_XC6eSyncPortA = 142,
	Automation1PsoWindowInput_XC6eSyncPortB = 143,
	Automation1PsoWindowInput_XC6eDrivePulseStream = 144,
	Automation1PsoWindowInput_XL5ePrimaryFeedback = 145,
	Automation1PsoWindowInput_XL5eAuxiliaryFeedback = 146,
	Automation1PsoWindowInput_XL5eSyncPortA = 147,
	Automation1PsoWindowInput_XL5eSyncPortB = 148,
	Automation1PsoWindowInput_XL5eDrivePulseStream = 149,
	Automation1PsoWindowInput_XC2PrimaryFeedback = 150,
	Automation1PsoWindowInput_XC2AuxiliaryFeedback = 151,
	Automation1PsoWindowInput_XC2DrivePulseStream = 152,
	Automation1PsoWindowInput_XC2ePrimaryFeedback = 153,
	Automation1PsoWindowInput_XC2eAuxiliaryFeedback = 154,
	Automation1PsoWindowInput_XC2eDrivePulseStream = 155,
	Automation1PsoWindowInput_XL2ePrimaryFeedback = 156,
	Automation1PsoWindowInput_XL2eAuxiliaryFeedback = 157,
	Automation1PsoWindowInput_XL2eSyncPortA = 158,
	Automation1PsoWindowInput_XL2eSyncPortB = 159,
	Automation1PsoWindowInput_XL2eDrivePulseStream = 160,
	Automation1PsoWindowInput_XI4PrimaryFeedbackAxis1 = 161,
	Automation1PsoWindowInput_XI4PrimaryFeedbackAxis2 = 162,
	Automation1PsoWindowInput_XI4PrimaryFeedbackAxis3 = 163,
	Automation1PsoWindowInput_XI4PrimaryFeedbackAxis4 = 164,
	Automation1PsoWindowInput_XI4AuxiliaryFeedback1 = 165,
	Automation1PsoWindowInput_XI4AuxiliaryFeedback2 = 166,
	Automation1PsoWindowInput_XI4AuxiliaryFeedback3 = 167,
	Automation1PsoWindowInput_XI4AuxiliaryFeedback4 = 168,
	Automation1PsoWindowInput_XI4SyncPortA = 169,
	Automation1PsoWindowInput_XI4SyncPortB = 170,
	Automation1PsoWindowInput_XI4DrivePulseStreamAxis1 = 171,
	Automation1PsoWindowInput_XI4DrivePulseStreamAxis2 = 172,
	Automation1PsoWindowInput_XI4DrivePulseStreamAxis3 = 173,
	Automation1PsoWindowInput_XI4DrivePulseStreamAxis4 = 174,
	Automation1PsoWindowInput_iXC4PrimaryFeedback = 175,
	Automation1PsoWindowInput_iXC4AuxiliaryFeedback = 176,
	Automation1PsoWindowInput_iXC4SyncPortA = 177,
	Automation1PsoWindowInput_iXC4SyncPortB = 178,
	Automation1PsoWindowInput_iXC4DrivePulseStream = 179,
	Automation1PsoWindowInput_iXC4ePrimaryFeedback = 180,
	Automation1PsoWindowInput_iXC4eAuxiliaryFeedback = 181,
	Automation1PsoWindowInput_iXC4eSyncPortA = 182,
	Automation1PsoWindowInput_iXC4eSyncPortB = 183,
	Automation1PsoWindowInput_iXC4eDrivePulseStream = 184,
	Automation1PsoWindowInput_iXC6ePrimaryFeedback = 185,
	Automation1PsoWindowInput_iXC6eAuxiliaryFeedback = 186,
	Automation1PsoWindowInput_iXC6eSyncPortA = 187,
	Automation1PsoWindowInput_iXC6eSyncPortB = 188,
	Automation1PsoWindowInput_iXC6eDrivePulseStream = 189,
	Automation1PsoWindowInput_iXL5ePrimaryFeedback = 190,
	Automation1PsoWindowInput_iXL5eAuxiliaryFeedback = 191,
	Automation1PsoWindowInput_iXL5eSyncPortA = 192,
	Automation1PsoWindowInput_iXL5eSyncPortB = 193,
	Automation1PsoWindowInput_iXL5eDrivePulseStream = 194,
	Automation1PsoWindowInput_iXR3PrimaryFeedbackAxis1 = 195,
	Automation1PsoWindowInput_iXR3PrimaryFeedbackAxis2 = 196,
	Automation1PsoWindowInput_iXR3PrimaryFeedbackAxis3 = 197,
	Automation1PsoWindowInput_iXR3PrimaryFeedbackAxis4 = 198,
	Automation1PsoWindowInput_iXR3PrimaryFeedbackAxis5 = 199,
	Automation1PsoWindowInput_iXR3PrimaryFeedbackAxis6 = 200,
	Automation1PsoWindowInput_iXR3AuxiliaryFeedbackAxis1 = 201,
	Automation1PsoWindowInput_iXR3AuxiliaryFeedbackAxis2 = 202,
	Automation1PsoWindowInput_iXR3AuxiliaryFeedbackAxis3 = 203,
	Automation1PsoWindowInput_iXR3AuxiliaryFeedbackAxis4 = 204,
	Automation1PsoWindowInput_iXR3AuxiliaryFeedbackAxis5 = 205,
	Automation1PsoWindowInput_iXR3AuxiliaryFeedbackAxis6 = 206,
	Automation1PsoWindowInput_iXR3SyncPortA = 207,
	Automation1PsoWindowInput_iXR3SyncPortB = 208,
	Automation1PsoWindowInput_iXR3DrivePulseStream = 209,
	Automation1PsoWindowInput_GI4DrivePulseStreamAxis1 = 210,
	Automation1PsoWindowInput_GI4DrivePulseStreamAxis2 = 211,
	Automation1PsoWindowInput_GI4DrivePulseStreamAxis3 = 212,
	Automation1PsoWindowInput_iXC2PrimaryFeedback = 213,
	Automation1PsoWindowInput_iXC2AuxiliaryFeedback = 214,
	Automation1PsoWindowInput_iXC2DrivePulseStream = 215,
	Automation1PsoWindowInput_iXC2ePrimaryFeedback = 216,
	Automation1PsoWindowInput_iXC2eAuxiliaryFeedback = 217,
	Automation1PsoWindowInput_iXC2eDrivePulseStream = 218,
	Automation1PsoWindowInput_iXL2ePrimaryFeedback = 219,
	Automation1PsoWindowInput_iXL2eAuxiliaryFeedback = 220,
	Automation1PsoWindowInput_iXL2eSyncPortA = 221,
	Automation1PsoWindowInput_iXL2eSyncPortB = 222,
	Automation1PsoWindowInput_iXL2eDrivePulseStream = 223,
	Automation1PsoWindowInput_iXI4PrimaryFeedbackAxis1 = 224,
	Automation1PsoWindowInput_iXI4PrimaryFeedbackAxis2 = 225,
	Automation1PsoWindowInput_iXI4PrimaryFeedbackAxis3 = 226,
	Automation1PsoWindowInput_iXI4PrimaryFeedbackAxis4 = 227,
	Automation1PsoWindowInput_iXI4AuxiliaryFeedback1 = 228,
	Automation1PsoWindowInput_iXI4AuxiliaryFeedback2 = 229,
	Automation1PsoWindowInput_iXI4AuxiliaryFeedback3 = 230,
	Automation1PsoWindowInput_iXI4AuxiliaryFeedback4 = 231,
	Automation1PsoWindowInput_iXI4SyncPortA = 232,
	Automation1PsoWindowInput_iXI4SyncPortB = 233,
	Automation1PsoWindowInput_iXI4DrivePulseStreamAxis1 = 234,
	Automation1PsoWindowInput_iXI4DrivePulseStreamAxis2 = 235,
	Automation1PsoWindowInput_iXI4DrivePulseStreamAxis3 = 236,
	Automation1PsoWindowInput_iXI4DrivePulseStreamAxis4 = 237,

# @brief Specifies the PSO output pin settings for each drive.
class Automation1PsoOutputPin:
	Automation1PsoOutputPin_GL4None = 100,
	Automation1PsoOutputPin_GL4LaserOutput0 = 101,
	Automation1PsoOutputPin_XL4sNone = 102,
	Automation1PsoOutputPin_XL4sLaserOutput0 = 103,
	Automation1PsoOutputPin_XR3None = 104,
	Automation1PsoOutputPin_XR3PsoOutput1 = 105,
	Automation1PsoOutputPin_XR3PsoOutput2 = 106,
	Automation1PsoOutputPin_XR3PsoOutput3 = 107,
	Automation1PsoOutputPin_XC4DedicatedOutput = 108,
	Automation1PsoOutputPin_XC4AuxiliaryMarkerDifferential = 109,
	Automation1PsoOutputPin_XC4AuxiliaryMarkerSingleEnded = 110,
	XC4eDedicatedOutput = 111,
	XC4eAuxiliaryMarkerDifferential = 112,
	XC4eAuxiliaryMarkerSingleEnded = 113,
	Automation1PsoOutputPin_XC6eDedicatedOutput = 114,
	Automation1PsoOutputPin_XC6eAuxiliaryMarkerDifferential = 115,
	Automation1PsoOutputPin_XC6eAuxiliaryMarkerSingleEnded = 116,
	Automation1PsoOutputPin_XL5eDedicatedOutput = 117,
	Automation1PsoOutputPin_XL5eAuxiliaryMarkerDifferential = 118,
	Automation1PsoOutputPin_XL5eAuxiliaryMarkerSingleEnded = 119,
	Automation1PsoOutputPin_XC2DedicatedOutput = 120,
	Automation1PsoOutputPin_XC2eDedicatedOutput = 121,
	Automation1PsoOutputPin_XL2eDedicatedOutput = 122,
	Automation1PsoOutputPin_XI4DedicatedOutput = 123,
	Automation1PsoOutputPin_iXC4DedicatedOutput = 124,
	Automation1PsoOutputPin_iXC4AuxiliaryMarkerDifferential = 125,
	Automation1PsoOutputPin_iXC4AuxiliaryMarkerSingleEnded = 126,
	Automation1PsoOutputPin_iXC4eDedicatedOutput = 127,
	Automation1PsoOutputPin_iXC4eAuxiliaryMarkerDifferential = 128,
	Automation1PsoOutputPin_iXC4eAuxiliaryMarkerSingleEnded = 129,
	Automation1PsoOutputPin_iXC6eDedicatedOutput = 130,
	Automation1PsoOutputPin_iXC6eAuxiliaryMarkerDifferential = 131,
	Automation1PsoOutputPin_iXC6eAuxiliaryMarkerSingleEnded = 132,
	Automation1PsoOutputPin_iXL5eDedicatedOutput = 133,
	Automation1PsoOutputPin_iXL5eAuxiliaryMarkerDifferential = 134,
	Automation1PsoOutputPin_iXL5eAuxiliaryMarkerSingleEnded = 135,
	Automation1PsoOutputPin_iXR3None = 136,
	Automation1PsoOutputPin_iXR3PsoOutput1 = 137,
	Automation1PsoOutputPin_iXR3PsoOutput2 = 138,
	Automation1PsoOutputPin_iXR3PsoOutput3 = 139,
	Automation1PsoOutputPin_GI4None = 140,
	Automation1PsoOutputPin_GI4LaserOutput0 = 141,
	Automation1PsoOutputPin_iXC2eDedicatedOutput = 143,
	Automation1PsoOutputPin_iXL2eDedicatedOutput = 144,
	Automation1PsoOutputPin_iXI4DedicatedOutput = 145,
"""


class DriveDataCaptureInput:
    PositionCommand = 0
    PrimaryFeedback = 1
    AuxiliaryFeedback = 2
    AnalogInput0 = 3
    AnalogInput1 = 4
    AnalogInput2 = 5
    AnalogInput3 = 6


class DriveDataCaptureTrigger:
    PsoOutput = 0
    PsoEvent = 1
    HighSpeedInput0RisingEdge = 2
    HighSpeedInput0FallingEdge = 3
    HighSpeedInput1RisingEdge = 4
    HighSpeedInput1FallingEdge = 5
    AuxiliaryMarkerRisingEdge = 6
    AuxiliaryMarkerFallingEdge = 7


class PsoOutputPin:
    GL4None = (100,)
    GL4LaserOutput0 = (101,)
    XL4sNone = (102,)
    XL4sLaserOutput0 = (103,)
    XR3None = (104,)
    XR3PsoOutput1 = (105,)
    XR3PsoOutput2 = (106,)
    XR3PsoOutput3 = (107,)
    XC4DedicatedOutput = (108,)
    XC4AuxiliaryMarkerDifferential = (109,)
    XC4AuxiliaryMarkerSingleEnded = (110,)
    XC4eDedicatedOutput = (111,)
    XC4eAuxiliaryMarkerDifferential = (112,)
    XC4eAuxiliaryMarkerSingleEnded = (113,)
    XC6eDedicatedOutput = (114,)
    XC6eAuxiliaryMarkerDifferential = (115,)
    XC6eAuxiliaryMarkerSingleEnded = (116,)
    XL5eDedicatedOutput = (117,)
    XL5eAuxiliaryMarkerDifferential = (118,)
    XL5eAuxiliaryMarkerSingleEnded = (119,)
    XC2DedicatedOutput = (120,)
    XC2eDedicatedOutput = (121,)
    XL2eDedicatedOutput = (122,)
    XI4DedicatedOutput = (123,)
    iXC4DedicatedOutput = (124,)
    iXC4AuxiliaryMarkerDifferential = (125,)
    iXC4AuxiliaryMarkerSingleEnded = (126,)
    iXC4eDedicatedOutput = (127,)
    iXC4eAuxiliaryMarkerDifferential = (128,)
    iXC4eAuxiliaryMarkerSingleEnded = (129,)
    iXC6eDedicatedOutput = (130,)
    iXC6eAuxiliaryMarkerDifferential = (131,)
    iXC6eAuxiliaryMarkerSingleEnded = (132,)
    iXL5eDedicatedOutput = (133,)
    iXL5eAuxiliaryMarkerDifferential = (134,)
    iXL5eAuxiliaryMarkerSingleEnded = (135,)
    iXR3None = (136,)
    iXR3PsoOutput1 = (137,)
    iXR3PsoOutput2 = (138,)
    iXR3PsoOutput3 = (139,)
    GI4None = (140,)
    GI4LaserOutput0 = (141,)
    iXC2DedicatedOutput = (142,)
    iXC2eDedicatedOutput = (143,)
    iXL2eDedicatedOutput = (144,)
    iXI4DedicatedOutput = (145,)



from unittest import mock

import pytest
from utils import SocketMock

from ophyd_devices.galil.fgalil_ophyd import FlomniGalilController, FlomniGalilMotor


@pytest.fixture(scope="function")
def leyey():
    FlomniGalilController._reset_controller()
    leyey_motor = FlomniGalilMotor(
        "H", name="leyey", host="mpc2680.psi.ch", port=8081, socket_cls=SocketMock
    )
    leyey_motor.controller.on()
    yield leyey_motor
    leyey_motor.controller.off()
    leyey_motor.controller._reset_controller()


@pytest.fixture(scope="function")
def leyex():
    FlomniGalilController._reset_controller()
    leyex_motor = FlomniGalilMotor(
        "H", name="leyey", host="mpc2680.psi.ch", port=8081, socket_cls=SocketMock
    )
    leyex_motor.controller.on()
    yield leyex_motor
    leyex_motor.controller.off()
    leyex_motor.controller._reset_controller()


@pytest.mark.parametrize("pos,msg,sign", [(1, b" -12800\n\r", 1), (-1, b" -12800\n\r", -1)])
def test_axis_get(leyey, pos, msg, sign):
    leyey.sign = sign
    leyey.controller.sock.flush_buffer()
    leyey.controller.sock.buffer_recv = msg
    val = leyey.read()
    assert val["leyey"]["value"] == pos
    assert leyey.readback.get() == pos


@pytest.mark.parametrize(
    "target_pos,socket_put_messages,socket_get_messages",
    [
        (
            0,
            [
                b"MG allaxref\r",
                b"MG_XQ0\r",
                b"naxis=7\r",
                b"ntarget=0.000\r",
                b"movereq=1\r",
                b"XQ#NEWPAR\r",
                b"MG_XQ0\r",
            ],
            [b"1.00", b"-1", b":", b":", b":", b":", b"-1"],
        )
    ],
)
def test_axis_put(leyey, target_pos, socket_put_messages, socket_get_messages):
    leyey.controller.sock.flush_buffer()
    leyey.controller.sock.buffer_recv = socket_get_messages
    leyey.user_setpoint.put(target_pos)
    assert leyey.controller.sock.buffer_put == socket_put_messages


@pytest.mark.parametrize(
    "axis_nr,direction,socket_put_messages,socket_get_messages",
    [
        (
            0,
            "forward",
            [
                b"naxis=0\r",
                b"ndir=1\r",
                b"XQ#NEWPAR\r",
                b"XQ#FES\r",
                b"MG_XQ0\r",
                b"MG _MOA\r",
                b"MG_XQ0\r",
                b"MG _MOA\r",
                b"MG _LRA, _LFA\r",
            ],
            [b":", b":", b":", b":", b"0", b"0", b"-1", b"-1", b"1.000 0.000"],
        ),
        (
            1,
            "reverse",
            [
                b"naxis=1\r",
                b"ndir=-1\r",
                b"XQ#NEWPAR\r",
                b"XQ#FES\r",
                b"MG_XQ0\r",
                b"MG _MOB\r",
                b"MG_XQ0\r",
                b"MG _MOB\r",
                b"MG _LRB, _LFB\r",
            ],
            [b":", b":", b":", b":", b"0", b"0", b"-1", b"-1", b"0.000 1.000"],
        ),
    ],
)
def test_drive_axis_to_limit(leyex, axis_nr, direction, socket_put_messages, socket_get_messages):
    leyex.controller.sock.flush_buffer()
    leyex.controller.sock.buffer_recv = socket_get_messages
    leyex.controller.drive_axis_to_limit(axis_nr, direction)
    assert leyex.controller.sock.buffer_put == socket_put_messages


@pytest.mark.parametrize(
    "axis_nr,socket_put_messages,socket_get_messages",
    [
        (
            0,
            [
                b"naxis=0\r",
                b"XQ#NEWPAR\r",
                b"XQ#FRM\r",
                b"MG_XQ0\r",
                b"MG _MOA\r",
                b"MG_XQ0\r",
                b"MG _MOA\r",
                b"MG axisref[0]\r",
            ],
            [b":", b":", b":", b"0", b"0", b"-1", b"-1", b"1.00"],
        ),
        (
            1,
            [
                b"naxis=1\r",
                b"XQ#NEWPAR\r",
                b"XQ#FRM\r",
                b"MG_XQ0\r",
                b"MG _MOB\r",
                b"MG_XQ0\r",
                b"MG _MOB\r",
                b"MG axisref[1]\r",
            ],
            [b":", b":", b":", b"0", b"0", b"-1", b"-1", b"1.00"],
        ),
    ],
)
def test_find_reference(leyex, axis_nr, socket_put_messages, socket_get_messages):
    leyex.controller.sock.flush_buffer()
    leyex.controller.sock.buffer_recv = socket_get_messages
    leyex.controller.find_reference(axis_nr)
    assert leyex.controller.sock.buffer_put == socket_put_messages


@pytest.mark.parametrize(
    "axis_Id,socket_put_messages,socket_get_messages,triggered",
    [
        ("A", [b"MG @IN[14]\r"], [b" 1.0000\n"], True),
        ("B", [b"MG @IN[14]\r"], [b" 0.0000\n"], False),
    ],
)
def test_fosaz_light_curtain_is_triggered(
    axis_Id, socket_put_messages, socket_get_messages, triggered
):
    """test that the light curtain is triggered"""
    fosaz = FlomniGalilMotor(
        axis_Id, name="fosaz", host="mpc2680.psi.ch", port=8081, socket_cls=SocketMock
    )
    fosaz.controller.on()
    fosaz.controller.sock.flush_buffer()
    fosaz.controller.sock.buffer_recv = socket_get_messages
    assert fosaz.controller.fosaz_light_curtain_is_triggered() == triggered
    assert fosaz.controller.sock.buffer_put == socket_put_messages
    fosaz.controller.off()
    fosaz.controller._reset_controller()



# pylint: skip-file
import os
import threading
from unittest import mock

import ophyd
import pytest
from bec_lib import MessageEndpoints, messages

from ophyd_devices.epics.devices.pilatus_csaxs import PilatuscSAXS
from tests.utils import DMMock, MockPV


def patch_dual_pvs(device):
    for walk in device.walk_signals():
        if not hasattr(walk.item, "_read_pv"):
            continue
        if not hasattr(walk.item, "_write_pv"):
            continue
        if walk.item._read_pv.pvname.endswith("_RBV"):
            walk.item._read_pv = walk.item._write_pv


@pytest.fixture(scope="function")
def mock_det():
    name = "pilatus"
    prefix = "X12SA-ES-PILATUS300K:"
    sim_mode = False
    dm = DMMock()
    with mock.patch.object(dm, "connector"):
        with (
            mock.patch("ophyd_devices.epics.devices.psi_detector_base.FileWriter"),
            mock.patch(
                "ophyd_devices.epics.devices.psi_detector_base.PSIDetectorBase._update_service_config"
            ),
        ):
            with mock.patch.object(ophyd, "cl") as mock_cl:
                mock_cl.get_pv = MockPV
                mock_cl.thread_class = threading.Thread
                with mock.patch.object(PilatuscSAXS, "_init"):
                    det = PilatuscSAXS(
                        name=name, prefix=prefix, device_manager=dm, sim_mode=sim_mode
                    )
                    patch_dual_pvs(det)
                    yield det


@pytest.mark.parametrize("trigger_source, detector_state", [(1, 0)])
# TODO rewrite this one, write test for init_detector, init_filewriter is tested
def test_init_detector(mock_det, trigger_source, detector_state):
    """Test the _init function:

    This includes testing the functions:
    - _init_detector
    - _stop_det
    - _set_trigger
    --> Testing the filewriter is done in test_init_filewriter

    Validation upon setting the correct PVs

    """
    mock_det.custom_prepare.initialize_detector()  # call the method you want to test
    assert mock_det.cam.acquire.get() == detector_state
    assert mock_det.cam.trigger_mode.get() == trigger_source


@pytest.mark.parametrize(
    "scaninfo, stopped, expected_exception",
    [
        (
            {
                "eacc": "e12345",
                "num_points": 500,
                "frames_per_trigger": 1,
                "filepath": "test.h5",
                "scan_id": "123",
                "mokev": 12.4,
            },
            False,
            False,
        ),
        (
            {
                "eacc": "e12345",
                "num_points": 500,
                "frames_per_trigger": 1,
                "filepath": "test.h5",
                "scan_id": "123",
                "mokev": 12.4,
            },
            True,
            False,
        ),
    ],
)
def test_stage(mock_det, scaninfo, stopped, expected_exception):
    with mock.patch.object(
        mock_det.custom_prepare, "publish_file_location"
    ) as mock_publish_file_location:
        mock_det.scaninfo.num_points = scaninfo["num_points"]
        mock_det.scaninfo.frames_per_trigger = scaninfo["frames_per_trigger"]
        mock_det.filewriter.compile_full_filename.return_value = scaninfo["filepath"]
        mock_det.device_manager.add_device("mokev", value=12.4)
        mock_det.stopped = stopped
        with (
            mock.patch.object(
                mock_det.custom_prepare, "prepare_detector_backend"
            ) as mock_data_backend,
            mock.patch.object(
                mock_det.custom_prepare, "update_readout_time"
            ) as mock_update_readout_time,
        ):
            mock_det.filepath = scaninfo["filepath"]
            if expected_exception:
                with pytest.raises(Exception):
                    mock_det.timeout = 0.1
                    mock_det.stage()
            else:
                mock_det.stage()
                mock_data_backend.assert_called_once()
                mock_update_readout_time.assert_called()
                # Check _prep_det
                assert mock_det.cam.num_images.get() == int(
                    scaninfo["num_points"] * scaninfo["frames_per_trigger"]
                )
                assert mock_det.cam.num_frames.get() == 1

                mock_publish_file_location.assert_called_with(done=False)


def test_pre_scan(mock_det):
    mock_det.custom_prepare.pre_scan()
    assert mock_det.cam.acquire.get() == 1


@pytest.mark.parametrize(
    "readout_time, expected_value", [(1e-3, 3e-3), (3e-3, 3e-3), (5e-3, 5e-3), (None, 3e-3)]
)
def test_update_readout_time(mock_det, readout_time, expected_value):
    if readout_time is None:
        mock_det.custom_prepare.update_readout_time()
        assert mock_det.readout_time == expected_value
    else:
        mock_det.scaninfo.readout_time = readout_time
        mock_det.custom_prepare.update_readout_time()
        assert mock_det.readout_time == expected_value


@pytest.mark.parametrize(
    "scaninfo",
    [
        (
            {
                "filepath": "test.h5",
                "filepath_raw": "test5_raw.h5",
                "successful": True,
                "done": False,
                "scan_id": "123",
            }
        ),
        (
            {
                "filepath": "test.h5",
                "filepath_raw": "test5_raw.h5",
                "successful": False,
                "done": True,
                "scan_id": "123",
            }
        ),
        (
            {
                "filepath": "test.h5",
                "filepath_raw": "test5_raw.h5",
                "successful": None,
                "done": True,
                "scan_id": "123",
            }
        ),
    ],
)
def test_publish_file_location(mock_det, scaninfo):
    mock_det.scaninfo.scan_id = scaninfo["scan_id"]
    mock_det.filepath = scaninfo["filepath"]
    mock_det.filepath_raw = scaninfo["filepath_raw"]
    mock_det.custom_prepare.publish_file_location(
        done=scaninfo["done"], successful=scaninfo["successful"]
    )
    if scaninfo["successful"] is None:
        msg = messages.FileMessage(
            file_path=scaninfo["filepath"],
            done=scaninfo["done"],
            metadata={"input_path": scaninfo["filepath_raw"]},
        )
    else:
        msg = messages.FileMessage(
            file_path=scaninfo["filepath"],
            done=scaninfo["done"],
            metadata={"input_path": scaninfo["filepath_raw"]},
            successful=scaninfo["successful"],
        )
    expected_calls = [
        mock.call(
            MessageEndpoints.public_file(scaninfo["scan_id"], mock_det.name),
            msg,
            pipe=mock_det.connector.pipeline.return_value,
        ),
        mock.call(
            MessageEndpoints.file_event(mock_det.name),
            msg,
            pipe=mock_det.connector.pipeline.return_value,
        ),
    ]
    assert mock_det.connector.set_and_publish.call_args_list == expected_calls


@pytest.mark.parametrize(
    "requests_state, expected_exception, url_delete, url_put",
    [
        (
            True,
            False,
            "http://x12sa-pd-2:8080/stream/pilatus_2",
            "http://xbl-daq-34:8091/pilatus_2/stop",
        ),
        (
            False,
            False,
            "http://x12sa-pd-2:8080/stream/pilatus_2",
            "http://xbl-daq-34:8091/pilatus_2/stop",
        ),
    ],
)
def test_stop_detector_backend(mock_det, requests_state, expected_exception, url_delete, url_put):
    with (
        mock.patch.object(
            mock_det.custom_prepare, "send_requests_delete"
        ) as mock_send_requests_delete,
        mock.patch.object(mock_det.custom_prepare, "send_requests_put") as mock_send_requests_put,
    ):
        instance_delete = mock_send_requests_delete.return_value
        instance_delete.ok = requests_state
        instance_put = mock_send_requests_put.return_value
        instance_put.ok = requests_state
        if expected_exception:
            mock_det.custom_prepare.stop_detector_backend()
            mock_send_requests_delete.assert_called_once_with(url=url_delete)
            mock_send_requests_put.assert_called_once_with(url=url_put)
            instance_delete.raise_for_status.called_once()
            instance_put.raise_for_status.called_once()
        else:
            mock_det.custom_prepare.stop_detector_backend()
            mock_send_requests_delete.assert_called_once_with(url=url_delete)
            mock_send_requests_put.assert_called_once_with(url=url_put)


@pytest.mark.parametrize(
    "scaninfo, data_msgs, urls, requests_state, expected_exception",
    [
        (
            {
                "filepath_raw": "pilatus_2.h5",
                "eacc": "e12345",
                "scan_number": 1000,
                "scan_directory": "S00000_00999",
                "num_points": 500,
                "frames_per_trigger": 1,
                "headers": {"Content-Type": "application/json", "Accept": "application/json"},
            },
            [
                {
                    "source": [
                        {
                            "searchPath": "/",
                            "searchPattern": "glob:*.cbf",
                            "destinationPath": (
                                "/sls/X12SA/data/e12345/Data10/pilatus_2/S00000_00999"
                            ),
                        }
                    ]
                },
                [
                    "zmqWriter",
                    "e12345",
                    {
                        "addr": "tcp://x12sa-pd-2:8888",
                        "dst": ["file"],
                        "numFrm": 500,
                        "timeout": 2000,
                        "ifType": "PULL",
                        "user": "e12345",
                    },
                ],
                ["zmqWriter", "e12345", {"frmCnt": 500, "timeout": 2000}],
            ],
            [
                "http://x12sa-pd-2:8080/stream/pilatus_2",
                "http://xbl-daq-34:8091/pilatus_2/run",
                "http://xbl-daq-34:8091/pilatus_2/wait",
            ],
            True,
            False,
        ),
        (
            {
                "filepath_raw": "pilatus_2.h5",
                "eacc": "e12345",
                "scan_number": 1000,
                "scan_directory": "S00000_00999",
                "num_points": 500,
                "frames_per_trigger": 1,
                "headers": {"Content-Type": "application/json", "Accept": "application/json"},
            },
            [
                {
                    "source": [
                        {
                            "searchPath": "/",
                            "searchPattern": "glob:*.cbf",
                            "destinationPath": (
                                "/sls/X12SA/data/e12345/Data10/pilatus_2/S00000_00999"
                            ),
                        }
                    ]
                },
                [
                    "zmqWriter",
                    "e12345",
                    {
                        "addr": "tcp://x12sa-pd-2:8888",
                        "dst": ["file"],
                        "numFrm": 500,
                        "timeout": 2000,
                        "ifType": "PULL",
                        "user": "e12345",
                    },
                ],
                ["zmqWriter", "e12345", {"frmCnt": 500, "timeout": 2000}],
            ],
            [
                "http://x12sa-pd-2:8080/stream/pilatus_2",
                "http://xbl-daq-34:8091/pilatus_2/run",
                "http://xbl-daq-34:8091/pilatus_2/wait",
            ],
            False,  # return of res.ok is False!
            True,
        ),
    ],
)
def test_prep_file_writer(mock_det, scaninfo, data_msgs, urls, requests_state, expected_exception):
    with (
        mock.patch.object(mock_det.custom_prepare, "close_file_writer") as mock_close_file_writer,
        mock.patch.object(mock_det.custom_prepare, "stop_file_writer") as mock_stop_file_writer,
        mock.patch.object(mock_det, "filewriter") as mock_filewriter,
        mock.patch.object(mock_det.custom_prepare, "create_directory") as mock_create_directory,
        mock.patch.object(mock_det.custom_prepare, "send_requests_put") as mock_send_requests_put,
    ):
        mock_det.scaninfo.scan_number = scaninfo["scan_number"]
        mock_det.scaninfo.num_points = scaninfo["num_points"]
        mock_det.scaninfo.frames_per_trigger = scaninfo["frames_per_trigger"]
        mock_det.scaninfo.username = scaninfo["eacc"]
        mock_filewriter.compile_full_filename.return_value = scaninfo["filepath_raw"]
        mock_filewriter.get_scan_directory.return_value = scaninfo["scan_directory"]
        instance = mock_send_requests_put.return_value
        instance.ok = requests_state
        instance.raise_for_status.side_effect = Exception

        if expected_exception:
            with pytest.raises(Exception):
                mock_det.timeout = 0.1
                mock_det.custom_prepare.prepare_data_backend()
                mock_close_file_writer.assert_called_once()
                mock_stop_file_writer.assert_called_once()
                instance.raise_for_status.assert_called_once()
        else:
            mock_det.custom_prepare.prepare_data_backend()

            mock_close_file_writer.assert_called_once()
            mock_stop_file_writer.assert_called_once()

            # Assert values set on detector
            assert mock_det.cam.file_path.get() == "/dev/shm/zmq/"
            assert (
                mock_det.cam.file_name.get()
                == f"{scaninfo['eacc']}_2_{scaninfo['scan_number']:05d}"
            )
            assert mock_det.cam.auto_increment.get() == 1
            assert mock_det.cam.file_number.get() == 0
            assert mock_det.cam.file_format.get() == 0
            assert mock_det.cam.file_template.get() == "%s%s_%5.5d.cbf"
            # Remove last / from destinationPath
            mock_create_directory.assert_called_once_with(
                os.path.join(data_msgs[0]["source"][0]["destinationPath"])
            )
            assert mock_send_requests_put.call_count == 3

            calls = [
                mock.call(url=url, data=data_msg, headers=scaninfo["headers"])
                for url, data_msg in zip(urls, data_msgs)
            ]
            for call, mock_call in zip(calls, mock_send_requests_put.call_args_list):
                assert call == mock_call


@pytest.mark.parametrize("stopped, expected_exception", [(False, False), (True, True)])
def test_unstage(mock_det, stopped, expected_exception):
    with (
        mock.patch.object(mock_det.custom_prepare, "finished") as mock_finished,
        mock.patch.object(
            mock_det.custom_prepare, "publish_file_location"
        ) as mock_publish_file_location,
    ):
        mock_det.stopped = stopped
        if expected_exception:
            mock_det.unstage()
            assert mock_det.stopped is True
        else:
            mock_det.unstage()
            mock_finished.assert_called_once()
            mock_publish_file_location.assert_called_with(done=True, successful=True)
            assert mock_det.stopped is False


def test_stop(mock_det):
    with (
        mock.patch.object(mock_det.custom_prepare, "stop_detector") as mock_stop_det,
        mock.patch.object(mock_det.custom_prepare, "stop_file_writer") as mock_stop_file_writer,
        mock.patch.object(mock_det.custom_prepare, "close_file_writer") as mock_close_file_writer,
    ):
        mock_det.stop()
        mock_stop_det.assert_called_once()
        mock_stop_file_writer.assert_called_once()
        mock_close_file_writer.assert_called_once()
        assert mock_det.stopped is True


@pytest.mark.parametrize(
    "stopped, mcs_stage_state, expected_exception",
    [
        (False, ophyd.Staged.no, False),
        (True, ophyd.Staged.no, True),
        (False, ophyd.Staged.yes, True),
    ],
)
def test_finished(mock_det, stopped, mcs_stage_state, expected_exception):
    with (
        mock.patch.object(mock_det, "device_manager") as mock_dm,
        mock.patch.object(mock_det.custom_prepare, "stop_file_writer") as mock_stop_file_friter,
        mock.patch.object(mock_det.custom_prepare, "stop_detector") as mock_stop_det,
        mock.patch.object(mock_det.custom_prepare, "close_file_writer") as mock_close_file_writer,
    ):
        mock_dm.devices.mcs.obj._staged = mcs_stage_state
        mock_det.stopped = stopped
        if expected_exception:
            with pytest.raises(Exception):
                mock_det.timeout = 0.1
                mock_det.custom_prepare.finished()
                assert mock_det.stopped is stopped
                mock_stop_file_friter.assert_called()
                mock_stop_det.assert_called_once()
                mock_close_file_writer.assert_called_once()
        else:
            mock_det.custom_prepare.finished()
            if stopped:
                assert mock_det.stopped is stopped

            mock_stop_file_friter.assert_called()
            mock_stop_det.assert_called_once()
            mock_close_file_writer.assert_called_once()



"""
ophyd device classes for X07MA beamline
"""

import time
import traceback
from collections import OrderedDict
from typing import Any

from bec_lib import bec_logger
from ophyd import Component as Cpt
from ophyd import Device, EpicsMotor, EpicsSignal, EpicsSignalRO
from ophyd import FormattedComponent as FCpt
from ophyd import Kind, PVPositioner, Signal
from ophyd.flyers import FlyerInterface
from ophyd.pv_positioner import PVPositionerComparator
from ophyd.status import DeviceStatus, SubscriptionStatus

logger = bec_logger.logger


__all__ = [
    "X07MAUndulator",
    "PGMMonochromator",
    "PGMOtFScan",
    "VacuumValve",
    "X07MAExitSlit",
    "X07MAMagnetAxis",
    "X07MAAnalogSignals",
    "X07MASampleManipulator",
    "X07MATemperatureController",
    "X07MAAutoTemperatureControl",
]


class X07MAUndulator(PVPositioner):
    """
    X07MA undulator
    """

    setpoint = Cpt(EpicsSignal, "ENERGY", auto_monitor=True)
    readback = Cpt(EpicsSignalRO, "ENERGY-READ", kind=Kind.hinted, auto_monitor=True)
    done = Cpt(EpicsSignalRO, "DONE", kind=Kind.omitted, auto_monitor=True)
    stop_signal = Cpt(EpicsSignal, "STOP", kind=Kind.omitted)

    energy_offset = Cpt(EpicsSignal, "ENERGY-OFFS", kind=Kind.config)
    pol_mode = Cpt(EpicsSignal, "MODE")
    pol_angle = Cpt(EpicsSignal, "ALPHA")
    harmonic = Cpt(EpicsSignal, "HARMONIC")

    def __init__(
        self,
        prefix="",
        *,
        limits=None,
        name=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        egu="",
        **kwargs,
    ):
        super().__init__(
            prefix,
            limits=limits,
            name=name,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            egu=egu,
            **kwargs,
        )
        self.readback.name = self.name


class PGMMonochromator(PVPositioner):
    """
    PGM monochromator
    """

    setpoint = Cpt(EpicsSignal, "PHS-E:GO.A", auto_monitor=True)
    readback = Cpt(EpicsSignalRO, "PGM:CERBK", kind=Kind.hinted, auto_monitor=True)
    done = Cpt(EpicsSignalRO, "PHS:alldone", kind=Kind.omitted, auto_monitor=True)
    stop_signal = Cpt(EpicsSignal, "PGM:stop", kind=Kind.omitted)

    cff = Cpt(EpicsSignal, "PGM:rbkcff", write_pv="PGM:cff.A", kind=Kind.config)
    with_undulator = Cpt(EpicsSignal, "PHS-E:OPT", kind=Kind.config)

    def __init__(
        self,
        prefix="",
        *,
        limits=None,
        name=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        egu="",
        **kwargs,
    ):
        super().__init__(
            prefix,
            limits=limits,
            name=name,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            egu=egu,
            **kwargs,
        )
        self.readback.name = self.name


class PGMOtFScan(FlyerInterface, Device):
    """
    PGM on-the-fly scan
    """

    SUB_VALUE = "value"
    SUB_FLYER = "flyer"
    _default_sub = SUB_VALUE

    e1 = Cpt(EpicsSignal, "E1", kind=Kind.config)
    e2 = Cpt(EpicsSignal, "E2", kind=Kind.config)
    time = Cpt(EpicsSignal, "TIME", kind=Kind.config)
    folder = Cpt(EpicsSignal, "FOLDER", kind=Kind.config)
    file = Cpt(EpicsSignal, "FILE", kind=Kind.config)
    acquire = Cpt(EpicsSignal, "START", auto_monitor=True)
    edata = Cpt(EpicsSignalRO, "EDATA", kind=Kind.hinted, auto_monitor=True)
    data = Cpt(EpicsSignalRO, "DATA", kind=Kind.hinted, auto_monitor=True)
    idata = Cpt(EpicsSignalRO, "IDATA", kind=Kind.hinted, auto_monitor=True)
    fdata = Cpt(EpicsSignalRO, "FDATA", kind=Kind.hinted, auto_monitor=True)
    count = Cpt(EpicsSignalRO, "COUNT", kind=Kind.omitted, auto_monitor=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._start_time = 0
        self.acquire.subscribe(self._update_status, run=False)
        self.count.subscribe(self._update_data, run=False)

    def kickoff(self):
        self._start_time = time.time()
        self.acquire.put(1, use_complete=True)
        status = DeviceStatus(self)
        status.set_finished()
        return status

    def complete(self):
        def check_value(*, old_value, value, **kwargs):
            return old_value == 1 and value == 0

        status = SubscriptionStatus(self.acquire, check_value, event_type=self.acquire.SUB_VALUE)
        return status

    def collect(self):
        data = {"time": self._start_time, "data": {}, "timestamps": {}}
        for attr in ("edata", "data", "idata", "fdata"):
            obj = getattr(self, attr)
            data["data"][obj.name] = obj.get()
            data["timestamps"][obj.name] = obj.timestamp

        return data

    def describe_collect(self):
        desc = OrderedDict()
        for attr in ("edata", "data", "idata", "fdata"):
            desc.update(getattr(self, attr).describe())
        return desc

    def _update_status(self, *, old_value, value, **kwargs):
        if old_value == 1 and value == 0:
            self._update_data(100)  # make sure that the last entries are also emitted
            self._done_acquiring()

    def _update_data(self, value, **kwargs):
        try:
            if value == 0:
                return
            data = self.collect()

            # FIXME: to avoid emitting outdated / stale data, wait until all signals report > 10 entries
            if any(len(val) < 10 for val in data["data"].values()) or value < 10:
                return
            self._run_subs(sub_type=self.SUB_FLYER, value=data)
        except Exception as exc:
            content = traceback.format_exc()
            logger.error(content)


class VacuumValve(PVPositionerComparator):
    """
    EPS vacuum valve.

    The setpoint is of 2 choices
       0 - Close
       1 - Try open

    The readback is of 8 choices
       0 - TO CONNECT
           1 - MAN OPEN
           2 - CLOSED
           3 - ERROR
           4 - MOVING
           5 - OPEN
           6 - ERROR
           7 - ERROR
    """

    setpoint = Cpt(EpicsSignal, "WT_SET")
    readback = Cpt(EpicsSignalRO, "POSITION")

    def __init__(self, prefix: str, *, name: str, **kwargs):
        kwargs.update({"limits": (0, 1)})
        super().__init__(prefix, name=name, **kwargs)
        self.readback.name = self.name

    def done_comparator(self, readback: Any, setpoint: Any) -> bool:
        return readback != 4


class X07MAExitSlit(PVPositioner):
    """
    Exit slit
    """

    setpoint = Cpt(EpicsSignal, "TR_AP")
    readback = Cpt(EpicsSignalRO, "TR_ISAP", kind=Kind.hinted, auto_monitor=True)
    done = Cpt(EpicsSignalRO, "TR.DMOV", kind=Kind.omitted, auto_monitor=True)

    def __init__(
        self,
        prefix="",
        *,
        limits=None,
        name=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        egu="",
        **kwargs,
    ):
        super().__init__(
            prefix,
            limits=limits,
            name=name,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            egu=egu,
            **kwargs,
        )
        self.readback.name = self.name


# class X07MAMagnet(Device):
#     """
#     Magnet fields.
#     """


class X07MAMagnetAxis(PVPositioner):
    """
    A single magnet field axis.
    """

    done_value = 1
    actuate_value = 1
    setpoint = FCpt(EpicsSignal, "{prefix}{_axis_id}:DMD")
    readback = FCpt(EpicsSignalRO, "{prefix}{_axis_id}:RBV", kind=Kind.hinted, auto_monitor=True)
    actuate = Cpt(EpicsSignal, "STARTRAMP.PROC", kind=Kind.omitted)
    done = FCpt(EpicsSignalRO, "{prefix}{_axis_id}:RAMP:DONE", auto_monitor=True)
    ramprate = FCpt(
        EpicsSignal, "{_ps_prefix}STS:RAMPRATE:TPM", write_pv="{_ps_prefix}SET:DMD:RAMPRATE:TPM"
    )

    def __init__(self, prefix="", axis_id="", ps_prefix="", *, name=None, **kwargs):
        self._axis_id = axis_id
        self._ps_prefix = ps_prefix
        super().__init__(prefix, name=name, **kwargs)
        self.readback.name = self.name

    def _setup_move(self, position):
        # If the setpoint is close to the current readback, the controller will not do anything,
        # and the done value will not change. So we simply say is done.
        if abs(position - self.readback.get()) < max(0.004, position * 0.0025):
            self._done_moving(success=True, time=time.ctime(), value=position)
        else:
            super()._setup_move(position)

    # x = Cpt(MagnetAxis, "", axis_id="X", ps_prefix="X07MA-PC-PS2:", name="x")
    # z = Cpt(MagnetAxis, "", axis_id="Z", ps_prefix="X07MA-PC-PS1:", name="z")


class NormSignal(Signal):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._metadata.update(write_access=False)

    def wait_for_connection(self, timeout=0):
        super().wait_for_connection(timeout)
        self._metadata.update(connected=True)

    def get(self, **kwargs):
        val1 = self.parent.signal1.get()
        val2 = self.parent.signal2.get()
        return val1 / val2 if val2 != 0 else 0

    def describe(self):
        desc = {
            "shape": [],
            "dtype": "number",
            "source": "PV: {} / {}".format(self.parent.signal1.pvname, self.parent.signal2.pvname),
            "units": "",
            "precision": self.parent.signal1.precision,
        }
        return desc


class NormTEYSignals(Device):
    signal1 = Cpt(EpicsSignalRO, name="signal1", read_pv="X07MA-ES1-AI:SIGNAL1", kind="omitted")
    signal2 = Cpt(EpicsSignalRO, name="signal2", read_pv="X07MA-ES1-AI:SIGNAL2", kind="omitted")
    norm = Cpt(NormSignal, name="norm", kind="hinted")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.norm.name = self.name


class NormDIODESignals(Device):
    signal1 = Cpt(EpicsSignalRO, name="signal1", read_pv="X07MA-ES1-AI:SIGNAL3", kind="omitted")
    signal2 = Cpt(EpicsSignalRO, name="signal2", read_pv="X07MA-ES1-AI:SIGNAL2", kind="omitted")
    norm = Cpt(NormSignal, name="norm", kind="hinted")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.norm.name = self.name


class X07MAAnalogSignals(Device):
    """
    ADC inputs
    """

    s1 = Cpt(EpicsSignalRO, "SIGNAL0", kind=Kind.hinted, auto_monitor=True)
    s2 = Cpt(EpicsSignalRO, "SIGNAL1", kind=Kind.hinted, auto_monitor=True)
    s3 = Cpt(EpicsSignalRO, "SIGNAL2", kind=Kind.hinted, auto_monitor=True)
    s4 = Cpt(EpicsSignalRO, "SIGNAL3", kind=Kind.hinted, auto_monitor=True)
    s5 = Cpt(EpicsSignalRO, "SIGNAL4", kind=Kind.hinted, auto_monitor=True)
    s6 = Cpt(EpicsSignalRO, "SIGNAL5", kind=Kind.hinted, auto_monitor=True)
    s7 = Cpt(EpicsSignalRO, "SIGNAL6", kind=Kind.hinted, auto_monitor=True)
    norm_tey = Cpt(NormTEYSignals, name="norm_tey", kind=Kind.hinted)
    norm_diode = Cpt(NormDIODESignals, name="norm_tey", kind=Kind.hinted)

    # Aliases
    # tey = s1
    # i0 = s2
    # trans = s3
    # field_z = s4
    # field_x = s5


class X07MASampleManipulator(Device):
    """
    Sample manipulator
    """

    hor = Cpt(EpicsMotor, "TRZS")
    vert = Cpt(EpicsMotor, "TRY1")
    rot = Cpt(EpicsMotor, "ROY1")


class X07MATemperatureController(Device):
    """
    Temperature controller
    """

    needle_valve = Cpt(EpicsSignal, "STS:LOOP2:MANUAL", write_pv="DMD:LOOP2:MANUAL")
    setpoint = Cpt(EpicsSignal, "STS:LOOP1:SETPOINT", write_pv="DMD:LOOP1:SETPOINT")
    readback = Cpt(EpicsSignalRO, "STS:T1", kind=Kind.hinted, auto_monitor=True)

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        **kwargs,
    ):
        super().__init__(
            prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        self.readback.name = self.name


class X07MAAutoTemperatureControl(Device):
    """
    Automatic temperature control.
    """

    control = Cpt(EpicsSignal, "CONTROL")
    status = Cpt(EpicsSignalRO, "STATUS", string=True, auto_monitor=True)



import functools
import socket
import threading
import time

from prettytable import PrettyTable
from typeguard import typechecked

from ophyd_devices.utils.controller import threadlocked
from ophyd_devices.utils.socket import raise_if_disconnected


def channel_checked(fcn):
    """Decorator to catch attempted access to channels that are not available."""

    @functools.wraps(fcn)
    def wrapper(self, *args, **kwargs):
        self._check_channel(args[0])
        return fcn(self, *args, **kwargs)

    return wrapper


class SocketIO:
    """SocketIO helper class for TCP IP connections"""

    def __init__(self, sock=None):
        self.is_open = False
        if sock is None:
            self.open()
        else:
            self.sock = sock

    def connect(self, host, port):
        print(f"connecting to {host} port {port}")
        # self.sock.create_connection((host, port))
        self.sock.connect((host, port))

    def _put(self, msg_bytes):
        return self.sock.send(msg_bytes)

    def _recv(self, buffer_length=1024):
        return self.sock.recv(buffer_length)

    def _initialize_socket(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.settimeout(5)

    def put(self, msg):
        return self._put(msg)

    def receive(self, buffer_length=1024):
        return self._recv(buffer_length=buffer_length)

    def open(self):
        self._initialize_socket()
        self.is_open = True

    def close(self):
        self.sock.close()
        self.sock = None
        self.is_open = False


class NPointController:
    _controller_instance = None

    NUM_CHANNELS = 3
    _read_single_loc_bit = "A0"
    _write_single_loc_bit = "A2"
    _trailing_bit = "55"
    _range_offset = "78"
    _channel_base = ["11", "83"]

    def __init__(
        self, comm_socket: SocketIO, server_ip: str = "129.129.99.87", server_port: int = 23
    ) -> None:
        self._lock = threading.RLock()
        super().__init__()
        self._server_and_port_name = (server_ip, server_port)
        self.socket = comm_socket
        self.connected = False

    def __new__(cls, *args, **kwargs):
        if not NPointController._controller_instance:
            NPointController._controller_instance = object.__new__(cls)
        return NPointController._controller_instance

    @classmethod
    def create(cls):
        return cls(SocketIO())

    def show_all(self) -> None:
        """Display current status of all channels

        Returns:
            None
        """
        if not self.connected:
            print("npoint controller is currently disabled.")
            return
        print(f"Connected to controller at {self._server_and_port_name}")
        t = PrettyTable()
        t.field_names = ["Channel", "Range", "Position", "Target"]
        for ii in range(self.NUM_CHANNELS):
            t.add_row(
                [ii, self._get_range(ii), self._get_current_pos(ii), self._get_target_pos(ii)]
            )
        print(t)

    @threadlocked
    def on(self) -> None:
        """Enable the NPoint controller and open a new socket.

        Raises:
            TimeoutError: Raised if the socket connection raises a timeout.

        Returns:
            None
        """
        if self.connected:
            print("You are already connected to the NPoint controller.")
            return
        if not self.socket.is_open:
            self.socket.open()
        try:
            self.socket.connect(self._server_and_port_name[0], self._server_and_port_name[1])
        except socket.timeout:
            raise TimeoutError(
                f"Failed to connect to the specified server and port {self._server_and_port_name}."
            )
        except OSError:
            print("ERROR while connecting. Let's try again")
            self.socket.close()
            time.sleep(0.5)
            self.socket.open()
            self.socket.connect(self._server_and_port_name[0], self._server_and_port_name[1])
        self.connected = True

    @threadlocked
    def off(self) -> None:
        """Disable the controller and close the socket.

        Returns:
            None
        """
        self.socket.close()
        self.connected = False

    @channel_checked
    def _get_range(self, channel: int) -> int:
        """Get the range of the specified channel axis.

        Args:
            channel (int): Channel for which the range should be requested.

        Raises:
            RuntimeError: Raised if the received message doesn't have the expected number of bytes (10).

        Returns:
            int: Range
        """

        # for first channel: 0x11 83 10 78
        addr = self._channel_base.copy()
        addr.extend([f"{16 + 16 * channel:x}", self._range_offset])
        send_buffer = self.__read_single_location_buffer(addr)

        recvd = self._put_and_receive(send_buffer)
        if len(recvd) != 10:
            raise RuntimeError(
                f"Received buffer is corrupted. Expected 10 bytes and instead got {len(recvd)}"
            )
        device_range = self._hex_list_to_int(recvd[5:-1], signed=False)
        return device_range

    @channel_checked
    def _get_current_pos(self, channel: int) -> float:
        # for first channel: 0x11 83 13 34
        addr = self._channel_base.copy()
        addr.extend([f"{19 + 16 * channel:x}", "34"])
        send_buffer = self.__read_single_location_buffer(addr)

        recvd = self._put_and_receive(send_buffer)

        pos_buffer = recvd[5:-1]
        pos = self._hex_list_to_int(pos_buffer) / 1048574 * 100
        return pos

    @channel_checked
    def _set_target_pos(self, channel: int, pos: float) -> None:
        # for first channel: 0x11 83 12 18 00 00 00 00
        addr = self._channel_base.copy()
        addr.extend([f"{18 + channel * 16:x}", "18"])

        target = int(round(1048574 / 100 * pos))
        data = [f"{m:02x}" for m in target.to_bytes(4, byteorder="big", signed=True)]

        send_buffer = self.__write_single_location_buffer(addr, data)
        self._put(send_buffer)

    @channel_checked
    def _get_target_pos(self, channel: int) -> float:
        # for first channel: 0x11 83 12 18
        addr = self._channel_base.copy()
        addr.extend([f"{18 + channel * 16:x}", "18"])
        send_buffer = self.__read_single_location_buffer(addr)

        recvd = self._put_and_receive(send_buffer)
        pos_buffer = recvd[5:-1]
        pos = self._hex_list_to_int(pos_buffer) / 1048574 * 100
        return pos

    @channel_checked
    def _set_servo(self, channel: int, enable: bool) -> None:
        print("Not tested")
        return
        # for first channel: 0x11 83 10 84 00 00 00 00
        addr = self._channel_base.copy()
        addr.extend([f"{16 + channel * 16:x}", "84"])

        if enable:
            data = ["00"] * 3 + ["01"]
        else:
            data = ["00"] * 4
        send_buffer = self.__write_single_location_buffer(addr, data)

        self._put(send_buffer)

    @channel_checked
    def _get_servo(self, channel: int) -> int:
        # for first channel: 0x11 83 10 84 00 00 00 00
        addr = self._channel_base.copy()
        addr.extend([f"{16 + channel * 16:x}", "84"])
        send_buffer = self.__read_single_location_buffer(addr)

        recvd = self._put_and_receive(send_buffer)
        buffer = recvd[5:-1]
        status = self._hex_list_to_int(buffer)
        return status

    @threadlocked
    def _put(self, buffer: list) -> None:
        """Translates a list of hex values to bytes and sends them to the socket.

        Args:
            buffer (list): List of hex values without leading 0x

        Returns:
            None
        """

        buffer = b"".join([bytes.fromhex(m) for m in buffer])
        self.socket.put(buffer)

    @threadlocked
    def _put_and_receive(self, msg_hex_list: list) -> list:
        """Send msg to socket and wait for a reply.

        Args:
            msg_hex_list (list): List of hex values without leading 0x.

        Returns:
            list: Received message as a list of hex values
        """

        buffer = b"".join([bytes.fromhex(m) for m in msg_hex_list])
        self.socket.put(buffer)
        recv_msg = self.socket.receive()
        recv_hex_list = [hex(m) for m in recv_msg]
        self._verify_received_msg(msg_hex_list, recv_hex_list)
        return recv_hex_list

    def _verify_received_msg(self, in_list: list, out_list: list) -> None:
        """Ensure that the first address bits of sent and received messages are the same.

        Args:
            in_list (list): list containing the sent message
            out_list (list): list containing the received message

        Raises:
            RuntimeError: Raised if first two address bits of 'in' and 'out' are not identical

        Returns:
            None
        """

        # first, translate hex (str) values to int
        in_list_int = [int(val, 16) for val in in_list]
        out_list_int = [int(val, 16) for val in out_list]

        # first ints of the reply should be the same. Otherwise something went wrong
        if not in_list_int[:2] == out_list_int[:2]:
            raise RuntimeError("Connection failure. Please restart the controller.")

    def _check_channel(self, channel: int) -> None:
        if channel >= self.NUM_CHANNELS:
            raise ValueError(
                f"Channel {channel+1} exceeds the available number of channels ({self.NUM_CHANNELS})"
            )

    @staticmethod
    def _hex_list_to_int(in_buffer: list, byteorder="little", signed=True) -> int:
        """Translate hex list to int.

        Args:
            in_buffer (list): Input buffer; received as list of hex values
            byteorder (str, optional): Byteorder of in_buffer. Defaults to "little".
            signed (bool, optional): Whether the hex list represents a signed int. Defaults to True.

        Returns:
            int: Translated integer.
        """
        if byteorder == "little":
            in_buffer.reverse()

        # make sure that all hex strings have the same format ("FF")
        val_hex = [f"{int(m, 16):02x}" for m in in_buffer]

        val_bytes = [bytes.fromhex(m) for m in val_hex]
        val = int.from_bytes(b"".join(val_bytes), byteorder="big", signed=signed)
        return val

    @staticmethod
    def __read_single_location_buffer(addr) -> list:
        """Prepare buffer for reading from a single memory location (hex address).
        Number of bytes: 6
        Format: 0xA0 [addr] 0x55
        Return Value: 0xA0 [addr] [data] 0x55
        Sample Hex Transmission from PC to LC.400: A0 18 12 83 11 55
        Sample Hex Return Transmission from LC.400 to PC: A0 18 12 83 11 64 00 00 00 55

        Args:
            addr (list): Hex address to read from

        Returns:
            list: List of hex values representing the read instruction.
        """
        buffer = []
        buffer.append(NPointController._read_single_loc_bit)
        if isinstance(addr, list):
            addr.reverse()
            buffer.extend(addr)
        else:
            buffer.append(addr)
        buffer.append(NPointController._trailing_bit)

        return buffer

    @staticmethod
    def __write_single_location_buffer(addr: list, data: list) -> list:
        """Prepare buffer for writing to a single memory location (hex address).
        Number of bytes: 10
        Format: 0xA2 [addr] [data] 0x55
        Return Value: none
        Sample Hex Transmission from PC to C.400: A2 18 12 83 11 E8 03 00 00 55

        Args:
            addr (list): List of hex values representing the address to write to.
            data (list): List of hex values representing the data that should be written.

        Returns:
            list: List of hex values representing the write instruction.
        """
        buffer = []
        buffer.append(NPointController._write_single_loc_bit)
        if isinstance(addr, list):
            addr.reverse()
            buffer.extend(addr)
        else:
            buffer.append(addr)

        if isinstance(data, list):
            data.reverse()
            buffer.extend(data)
        else:
            buffer.append(data)
        buffer.append(NPointController._trailing_bit)
        return buffer

    @staticmethod
    def __read_array():
        raise NotImplementedError

    @staticmethod
    def __write_next_command():
        raise NotImplementedError

    def __del__(self):
        if self.connected:
            print("Closing npoint socket")
            self.off()


class NPointAxis:
    def __init__(self, controller: NPointController, channel: int, name: str) -> None:
        super().__init__()
        self._axis_range = 100
        self.controller = controller
        self.channel = channel
        self.name = name
        self.controller._check_channel(channel)
        self._settling_time = 0.1

        if self.settling_time == 0:
            self.settling_time = 0.1
            print(f"Setting the npoint settling time to {self.settling_time:.2f} s.")
            print(
                "You can set the settling time depending on the stage tuning\nusing the settling_time property."
            )
            print("This is the waiting time before the counting is done.")

    def show_all(self) -> None:
        self.controller.show_all()

    @raise_if_disconnected
    def get(self) -> float:
        """Get current position for this channel.

        Raises:
            RuntimeError: Raised if channel is not connected.

        Returns:
            float: position
        """
        return self.controller._get_current_pos(self.channel)

    @raise_if_disconnected
    def get_target_pos(self) -> float:
        """Get target position for this channel.

        Raises:
            RuntimeError: Raised if channel is not connected.

        Returns:
            float: position
        """
        return self.controller._get_target_pos(self.channel)

    @raise_if_disconnected
    @typechecked
    def set(self, pos: float) -> None:
        """Set a new target position and wait until settled (settling_time).

        Args:
            pos (float): New target position

        Raises:
            RuntimeError: Raised if channel is not connected.

        Returns:
            None
        """
        self.controller._set_target_pos(self.channel, pos)
        time.sleep(self.settling_time)

    @property
    def connected(self) -> bool:
        return self.controller.connected

    @property
    @raise_if_disconnected
    def servo(self) -> int:
        """Get servo status

        Raises:
            RuntimeError: Raised if channel is not connected.

        Returns:
            int: Servo status
        """
        return self.controller._get_servo(self.channel)

    @servo.setter
    @raise_if_disconnected
    @typechecked
    def servo(self, val: bool) -> None:
        """Set servo status

        Args:
            val (bool): Servo status

        Raises:
            RuntimeError: Raised if channel is not connected.

        Returns:
            None
        """
        self.controller._set_servo(self.channel, val)

    @property
    def settling_time(self) -> float:
        return self._settling_time

    @settling_time.setter
    @typechecked
    def settling_time(self, val: float) -> None:
        self._settling_time = val
        print(f"Setting the npoint settling time to {val:.2f} s.")


class NPointEpics(NPointAxis):
    def __init__(self, controller: NPointController, channel: int, name: str) -> None:
        super().__init__(controller, channel, name)
        self.low_limit = -50
        self.high_limit = 50
        self._prefix = name

    def get_pv(self) -> str:
        return self.name

    def get_position(self, readback=True) -> float:
        if readback:
            return self.get()
        else:
            return self.get_target_pos()

    def within_limits(self, pos: float) -> bool:
        return pos > self.low_limit and pos < self.high_limit

    def move(self, position: float, wait=True) -> None:
        self.set(position)


if __name__ == "__main__":
    ## EXAMPLES ##
    #
    # Create controller and socket instance explicitly:
    #     controller = NPointController(SocketIO())
    #     npointx = NPointAxis(controller, 0, "nx")
    #     npointy = NPointAxis(controller, 1, "ny")

    # Create controller instance explicitly
    #     controller = NPointController.create()
    #     npointx = NPointAxis(controller, 0, "nx")
    #     npointy = NPointAxis(controller, 1, "ny")

    # Single-line axis:
    #     npointx = NPointAxis(NPointController.create(), 0, "nx")
    #
    # EPICS wrapper:
    #     nx = NPointEpics(NPointController.create(), 0, "nx")

    controller = NPointController.create()
    npointx = NPointAxis(NPointController.create(), 0, "nx")
    npointy = NPointAxis(NPointController.create(), 0, "ny")



import enum
import os

from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from ophyd import Component as Cpt
from ophyd import Device, EpicsSignal, EpicsSignalRO, EpicsSignalWithRBV
from ophyd.mca import EpicsMCARecord

from ophyd_devices.epics.devices.psi_detector_base import CustomDetectorMixin, PSIDetectorBase

logger = bec_logger.logger


class FalconError(Exception):
    """Base class for exceptions in this module."""


class FalconTimeoutError(FalconError):
    """Raised when the Falcon does not respond in time."""


class DetectorState(enum.IntEnum):
    """Detector states for Falcon detector"""

    DONE = 0
    ACQUIRING = 1


class TriggerSource(enum.IntEnum):
    """Trigger source for Falcon detector"""

    USER = 0
    GATE = 1
    SYNC = 2


class MappingSource(enum.IntEnum):
    """Mapping source for Falcon detector"""

    SPECTRUM = 0
    MAPPING = 1


class EpicsDXPFalcon(Device):
    """
    DXP parameters for Falcon detector

    Base class to map EPICS PVs from DXP parameters to ophyd signals.
    """

    elapsed_live_time = Cpt(EpicsSignal, "ElapsedLiveTime")
    elapsed_real_time = Cpt(EpicsSignal, "ElapsedRealTime")
    elapsed_trigger_live_time = Cpt(EpicsSignal, "ElapsedTriggerLiveTime")

    # Energy Filter PVs
    energy_threshold = Cpt(EpicsSignalWithRBV, "DetectionThreshold")
    min_pulse_separation = Cpt(EpicsSignalWithRBV, "MinPulsePairSeparation")
    detection_filter = Cpt(EpicsSignalWithRBV, "DetectionFilter", string=True)
    scale_factor = Cpt(EpicsSignalWithRBV, "ScaleFactor")
    risetime_optimisation = Cpt(EpicsSignalWithRBV, "RisetimeOptimization")

    # Misc PVs
    detector_polarity = Cpt(EpicsSignalWithRBV, "DetectorPolarity")
    decay_time = Cpt(EpicsSignalWithRBV, "DecayTime")

    current_pixel = Cpt(EpicsSignalRO, "CurrentPixel")


class FalconHDF5Plugins(Device):
    """
    HDF5 parameters for Falcon detector

    Base class to map EPICS PVs from HDF5 Plugin to ophyd signals.
    """

    capture = Cpt(EpicsSignalWithRBV, "Capture")
    enable = Cpt(EpicsSignalWithRBV, "EnableCallbacks", string=True, kind="config")
    xml_file_name = Cpt(EpicsSignalWithRBV, "XMLFileName", string=True, kind="config")
    lazy_open = Cpt(EpicsSignalWithRBV, "LazyOpen", string=True, doc="0='No' 1='Yes'")
    temp_suffix = Cpt(EpicsSignalWithRBV, "TempSuffix", string=True)
    file_path = Cpt(EpicsSignalWithRBV, "FilePath", string=True, kind="config")
    file_name = Cpt(EpicsSignalWithRBV, "FileName", string=True, kind="config")
    file_template = Cpt(EpicsSignalWithRBV, "FileTemplate", string=True, kind="config")
    num_capture = Cpt(EpicsSignalWithRBV, "NumCapture", kind="config")
    file_write_mode = Cpt(EpicsSignalWithRBV, "FileWriteMode", kind="config")
    queue_size = Cpt(EpicsSignalWithRBV, "QueueSize", kind="config")
    array_counter = Cpt(EpicsSignalWithRBV, "ArrayCounter", kind="config")


class FalconSetup(CustomDetectorMixin):
    """
    Falcon setup class for cSAXS

    Parent class: CustomDetectorMixin

    """

    def initialize_default_parameter(self) -> None:
        """
        Set default parameters for Falcon

        This will set:
        - readout (float): readout time in seconds
        - value_pixel_per_buffer (int): number of spectra in buffer of Falcon Sitoro

        """
        self.parent.value_pixel_per_buffer = 20
        self.update_readout_time()

    def update_readout_time(self) -> None:
        """Set readout time for Eiger9M detector"""
        readout_time = (
            self.parent.scaninfo.readout_time
            if hasattr(self.parent.scaninfo, "readout_time")
            else self.parent.MIN_READOUT
        )
        self.parent.readout_time = max(readout_time, self.parent.MIN_READOUT)

    def initialize_detector(self) -> None:
        """Initialize Falcon detector"""
        self.stop_detector()
        self.stop_detector_backend()
        self.parent.set_trigger(
            mapping_mode=MappingSource.MAPPING, trigger_source=TriggerSource.GATE, ignore_gate=0
        )
        # 1 Realtime
        self.parent.preset_mode.put(1)
        # 0 Normal, 1 Inverted
        self.parent.input_logic_polarity.put(0)
        # 0 Manual 1 Auto
        self.parent.auto_pixels_per_buffer.put(0)
        # Sets the number of pixels/spectra in the buffer
        self.parent.pixels_per_buffer.put(self.parent.value_pixel_per_buffer)

    def stop_detector(self) -> None:
        """Stops detector"""

        self.parent.stop_all.put(1)
        self.parent.erase_all.put(1)

        signal_conditions = [
            (lambda: self.parent.state.read()[self.parent.state.name]["value"], DetectorState.DONE)
        ]

        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout - self.parent.timeout // 2,
            all_signals=False,
        ):
            # Retry stop detector and wait for remaining time
            raise FalconTimeoutError(
                f"Failed to stop detector, timeout with state {signal_conditions[0][0]}"
            )

    def stop_detector_backend(self) -> None:
        """Stop the detector backend"""
        self.parent.hdf5.capture.put(0)

    def initialize_detector_backend(self) -> None:
        """Initialize the detector backend for Falcon."""
        self.parent.hdf5.enable.put(1)
        # file location of h5 layout for cSAXS
        self.parent.hdf5.xml_file_name.put("layout.xml")
        # TODO Check if lazy open is needed and wanted!
        self.parent.hdf5.lazy_open.put(1)
        self.parent.hdf5.temp_suffix.put("")
        # size of queue for number of spectra allowed in the buffer, if too small at high throughput, data is lost
        self.parent.hdf5.queue_size.put(2000)
        # Segmentation into Spectra within EPICS, 1 is activate, 0 is deactivate
        self.parent.nd_array_mode.put(1)

    def prepare_detector(self) -> None:
        """Prepare detector for acquisition"""
        self.parent.set_trigger(
            mapping_mode=MappingSource.MAPPING, trigger_source=TriggerSource.GATE, ignore_gate=0
        )
        self.parent.preset_real.put(self.parent.scaninfo.exp_time)
        self.parent.pixels_per_run.put(
            int(self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger)
        )

    def prepare_data_backend(self) -> None:
        """Prepare data backend for acquisition"""
        self.parent.filepath = self.parent.filewriter.compile_full_filename(
            f"{self.parent.name}.h5"
        )
        file_path, file_name = os.path.split(self.parent.filepath)
        self.parent.hdf5.file_path.put(file_path)
        self.parent.hdf5.file_name.put(file_name)
        self.parent.hdf5.file_template.put("%s%s")
        self.parent.hdf5.num_capture.put(
            int(self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger)
        )
        self.parent.hdf5.file_write_mode.put(2)
        # Reset spectrum counter in filewriter, used for indexing & identifying missing triggers
        self.parent.hdf5.array_counter.put(0)
        # Start file writing
        self.parent.hdf5.capture.put(1)

    def arm_acquisition(self) -> None:
        """Arm detector for acquisition"""
        self.parent.start_all.put(1)
        signal_conditions = [
            (
                lambda: self.parent.state.read()[self.parent.state.name]["value"],
                DetectorState.ACQUIRING,
            )
        ]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout,
            check_stopped=True,
            all_signals=False,
        ):
            raise FalconTimeoutError(
                f"Failed to arm the acquisition. Detector state {signal_conditions[0][0]}"
            )

    def check_scan_id(self) -> None:
        """Checks if scan_id has changed and stops the scan if it has"""
        old_scan_id = self.parent.scaninfo.scan_id
        self.parent.scaninfo.load_scan_metadata()
        if self.parent.scaninfo.scan_id != old_scan_id:
            self.parent.stopped = True

    def publish_file_location(self, done: bool = False, successful: bool = None) -> None:
        """
        Publish the filepath to REDIS.

        We publish two events here:
        - file_event: event for the filewriter
        - public_file: event for any secondary service (e.g. radial integ code)

        Args:
            done (bool): True if scan is finished
            successful (bool): True if scan was successful
        """
        pipe = self.parent.connector.pipeline()
        if successful is None:
            msg = messages.FileMessage(file_path=self.parent.filepath, done=done)
        else:
            msg = messages.FileMessage(
                file_path=self.parent.filepath, done=done, successful=successful
            )
        self.parent.connector.set_and_publish(
            MessageEndpoints.public_file(self.parent.scaninfo.scan_id, self.parent.name),
            msg,
            pipe=pipe,
        )
        self.parent.connector.set_and_publish(
            MessageEndpoints.file_event(self.parent.name), msg, pipe=pipe
        )
        pipe.execute()

    def finished(self) -> None:
        """Check if scan finished succesfully"""
        total_frames = int(
            self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger
        )
        signal_conditions = [
            (self.parent.dxp.current_pixel.get, total_frames),
            (self.parent.hdf5.array_counter.get, total_frames),
        ]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout,
            check_stopped=True,
            all_signals=True,
        ):
            logger.debug(
                f"Falcon missed a trigger: received trigger {self.parent.dxp.current_pixel.get()},"
                f" send data {self.parent.hdf5.array_counter.get()} from total_frames"
                f" {total_frames}"
            )
        self.stop_detector()
        self.stop_detector_backend()


class FalconcSAXS(PSIDetectorBase):
    """
    Falcon Sitoro detector for CSAXS

    Parent class: PSIDetectorBase

    class attributes:
        custom_prepare_cls (FalconSetup)        : Custom detector setup class for cSAXS,
                                                  inherits from CustomDetectorMixin
        PSIDetectorBase.set_min_readout (float) : Minimum readout time for the detector
        dxp (EpicsDXPFalcon)                    : DXP parameters for Falcon detector
        mca (EpicsMCARecord)                    : MCA parameters for Falcon detector
        hdf5 (FalconHDF5Plugins)                : HDF5 parameters for Falcon detector
        MIN_READOUT (float)                     : Minimum readout time for the detector
    """

    # Specify which functions are revealed to the user in BEC client
    USER_ACCESS = ["describe"]

    # specify Setup class
    custom_prepare_cls = FalconSetup
    # specify minimum readout time for detector
    MIN_READOUT = 3e-3

    # specify class attributes
    dxp = Cpt(EpicsDXPFalcon, "dxp1:")
    mca = Cpt(EpicsMCARecord, "mca1")
    hdf5 = Cpt(FalconHDF5Plugins, "HDF1:")

    stop_all = Cpt(EpicsSignal, "StopAll")
    erase_all = Cpt(EpicsSignal, "EraseAll")
    start_all = Cpt(EpicsSignal, "StartAll")
    state = Cpt(EpicsSignal, "Acquiring")
    preset_mode = Cpt(EpicsSignal, "PresetMode")  # 0 No preset 1 Real time 2 Events 3 Triggers
    preset_real = Cpt(EpicsSignal, "PresetReal")
    preset_events = Cpt(EpicsSignal, "PresetEvents")
    preset_triggers = Cpt(EpicsSignal, "PresetTriggers")
    triggers = Cpt(EpicsSignalRO, "MaxTriggers", lazy=True)
    events = Cpt(EpicsSignalRO, "MaxEvents", lazy=True)
    input_count_rate = Cpt(EpicsSignalRO, "MaxInputCountRate", lazy=True)
    output_count_rate = Cpt(EpicsSignalRO, "MaxOutputCountRate", lazy=True)
    collect_mode = Cpt(EpicsSignal, "CollectMode")  # 0 MCA spectra, 1 MCA mapping
    pixel_advance_mode = Cpt(EpicsSignal, "PixelAdvanceMode")
    ignore_gate = Cpt(EpicsSignal, "IgnoreGate")
    input_logic_polarity = Cpt(EpicsSignal, "InputLogicPolarity")
    auto_pixels_per_buffer = Cpt(EpicsSignal, "AutoPixelsPerBuffer")
    pixels_per_buffer = Cpt(EpicsSignal, "PixelsPerBuffer")
    pixels_per_run = Cpt(EpicsSignal, "PixelsPerRun")
    nd_array_mode = Cpt(EpicsSignal, "NDArrayMode")

    def set_trigger(
        self, mapping_mode: MappingSource, trigger_source: TriggerSource, ignore_gate: int = 0
    ) -> None:
        """
        Set triggering mode for detector

        Args:
            mapping_mode (MappingSource): Mapping mode for the detector
            trigger_source (TriggerSource): Trigger source for the detector, pixel_advance_signal
            ignore_gate (int): Ignore gate from TTL signal; defaults to 0

        """
        mapping = int(mapping_mode)
        trigger = trigger_source
        self.collect_mode.put(mapping)
        self.pixel_advance_mode.put(trigger)
        self.ignore_gate.put(ignore_gate)

    def stage(self) -> list[object]:
        """Stage"""
        rtr = super().stage()
        self.custom_prepare.arm_acquisition()
        return rtr


if __name__ == "__main__":
    falcon = FalconcSAXS(name="falcon", prefix="X12SA-SITORO:", sim_mode=True)



TABLES_DT_PUSH_DIST_MM = 890


class DetectorTableTheta(PseudoPositioner):
    """Detector table tilt motor

    Small wrapper to adjust the detector table tilt as angle.
    The table is pushed from one side by a single vertical motor.

    Note: Rarely used!
    """

    # Real axis (in degrees)
    pusher = Component(EpicsMotor, "", name="pusher")
    # Virtual axis
    theta = Component(PseudoSingle, name="theta")

    _real = ["pusher"]

    @pseudo_position_argument
    def forward(self, pseudo_pos):
        return self.RealPosition(
            pusher=tan(pseudo_pos.theta * 3.141592 / 180.0) * TABLES_DT_PUSH_DIST_MM
        )

    @real_position_argument
    def inverse(self, real_pos):
        return self.PseudoPosition(
            theta=-180 * atan(real_pos.pusher / TABLES_DT_PUSH_DIST_MM) / 3.141592
        )



# pylint: skip-file
import os
import threading
from unittest import mock

import ophyd
import pytest
from bec_lib import MessageEndpoints, messages

from ophyd_devices.epics.devices.falcon_csaxs import FalconcSAXS, FalconTimeoutError
from tests.utils import DMMock, MockPV


def patch_dual_pvs(device):
    for walk in device.walk_signals():
        if not hasattr(walk.item, "_read_pv"):
            continue
        if not hasattr(walk.item, "_write_pv"):
            continue
        if walk.item._read_pv.pvname.endswith("_RBV"):
            walk.item._read_pv = walk.item._write_pv


@pytest.fixture(scope="function")
def mock_det():
    name = "falcon"
    prefix = "X12SA-SITORO:"
    sim_mode = False
    dm = DMMock()
    with mock.patch.object(dm, "connector"):
        with (
            mock.patch("ophyd_devices.epics.devices.psi_detector_base.FileWriter") as filemixin,
            mock.patch(
                "ophyd_devices.epics.devices.psi_detector_base.PSIDetectorBase._update_service_config"
            ) as mock_service_config,
        ):
            with mock.patch.object(ophyd, "cl") as mock_cl:
                mock_cl.get_pv = MockPV
                mock_cl.thread_class = threading.Thread
                with mock.patch.object(FalconcSAXS, "_init"):
                    det = FalconcSAXS(
                        name=name, prefix=prefix, device_manager=dm, sim_mode=sim_mode
                    )
                    patch_dual_pvs(det)
                    yield det


@pytest.mark.parametrize(
    "trigger_source, mapping_source, ignore_gate, pixels_per_buffer, detector_state,"
    " expected_exception",
    [(1, 1, 0, 20, 0, False), (1, 1, 0, 20, 1, True)],
)
# TODO rewrite this one, write test for init_detector, init_filewriter is tested
def test_init_detector(
    mock_det,
    trigger_source,
    mapping_source,
    ignore_gate,
    pixels_per_buffer,
    detector_state,
    expected_exception,
):
    """Test the _init function:

    This includes testing the functions:
    - _init_detector
    - _stop_det
    - _set_trigger
    --> Testing the filewriter is done in test_init_filewriter

    Validation upon setting the correct PVs

    """
    mock_det.value_pixel_per_buffer = pixels_per_buffer
    mock_det.state._read_pv.mock_data = detector_state
    if expected_exception:
        with pytest.raises(FalconTimeoutError):
            mock_det.timeout = 0.1
            mock_det.custom_prepare.initialize_detector()
    else:
        mock_det.custom_prepare.initialize_detector()
        assert mock_det.state.get() == detector_state
        assert mock_det.collect_mode.get() == mapping_source
        assert mock_det.pixel_advance_mode.get() == trigger_source
        assert mock_det.ignore_gate.get() == ignore_gate

        assert mock_det.preset_mode.get() == 1
        assert mock_det.erase_all.get() == 1
        assert mock_det.input_logic_polarity.get() == 0
        assert mock_det.auto_pixels_per_buffer.get() == 0
        assert mock_det.pixels_per_buffer.get() == pixels_per_buffer


@pytest.mark.parametrize(
    "readout_time, expected_value", [(1e-3, 3e-3), (3e-3, 3e-3), (5e-3, 5e-3), (None, 3e-3)]
)
def test_update_readout_time(mock_det, readout_time, expected_value):
    if readout_time is None:
        mock_det.custom_prepare.update_readout_time()
        assert mock_det.readout_time == expected_value
    else:
        mock_det.scaninfo.readout_time = readout_time
        mock_det.custom_prepare.update_readout_time()
        assert mock_det.readout_time == expected_value


def test_initialize_default_parameter(mock_det):
    with mock.patch.object(
        mock_det.custom_prepare, "update_readout_time"
    ) as mock_update_readout_time:
        mock_det.custom_prepare.initialize_default_parameter()
        assert mock_det.value_pixel_per_buffer == 20
        mock_update_readout_time.assert_called_once()


@pytest.mark.parametrize(
    "scaninfo",
    [
        (
            {
                "eacc": "e12345",
                "num_points": 500,
                "frames_per_trigger": 1,
                "exp_time": 0.1,
                "filepath": "test.h5",
                "scan_id": "123",
                "mokev": 12.4,
            }
        )
    ],
)
def test_stage(mock_det, scaninfo):
    """Test the stage function:

    This includes testing _prep_det
    """
    with (
        mock.patch.object(mock_det, "set_trigger") as mock_set_trigger,
        mock.patch.object(
            mock_det.custom_prepare, "prepare_detector_backend"
        ) as mock_prep_data_backend,
        mock.patch.object(
            mock_det.custom_prepare, "publish_file_location"
        ) as mock_publish_file_location,
        mock.patch.object(mock_det.custom_prepare, "arm_acquisition") as mock_arm_acquisition,
    ):
        mock_det.scaninfo.exp_time = scaninfo["exp_time"]
        mock_det.scaninfo.num_points = scaninfo["num_points"]
        mock_det.scaninfo.frames_per_trigger = scaninfo["frames_per_trigger"]
        mock_det.stage()
        mock_set_trigger.assert_called_once()
        assert mock_det.preset_real.get() == scaninfo["exp_time"]
        assert mock_det.pixels_per_run.get() == int(
            scaninfo["num_points"] * scaninfo["frames_per_trigger"]
        )
        mock_prep_data_backend.assert_called_once()
        mock_publish_file_location.assert_called_once_with(done=False)
        mock_arm_acquisition.assert_called_once()


@pytest.mark.parametrize(
    "scaninfo",
    [
        (
            {
                "filepath": "/das/work/p18/p18533/data/S00000-S00999/S00001/data.h5",
                "num_points": 500,
                "frames_per_trigger": 1,
            }
        ),
        (
            {
                "filepath": "/das/work/p18/p18533/data/S00000-S00999/S00001/data1234.h5",
                "num_points": 500,
                "frames_per_trigger": 1,
            }
        ),
    ],
)
def test_prepare_data_backend(mock_det, scaninfo):
    mock_det.filewriter.compile_full_filename.return_value = scaninfo["filepath"]
    mock_det.scaninfo.num_points = scaninfo["num_points"]
    mock_det.scaninfo.frames_per_trigger = scaninfo["frames_per_trigger"]
    mock_det.scaninfo.scan_number = 1
    mock_det.custom_prepare.prepare_data_backend()
    file_path, file_name = os.path.split(scaninfo["filepath"])
    assert mock_det.hdf5.file_path.get() == file_path
    assert mock_det.hdf5.file_name.get() == file_name
    assert mock_det.hdf5.file_template.get() == "%s%s"
    assert mock_det.hdf5.num_capture.get() == int(
        scaninfo["num_points"] * scaninfo["frames_per_trigger"]
    )
    assert mock_det.hdf5.file_write_mode.get() == 2
    assert mock_det.hdf5.array_counter.get() == 0
    assert mock_det.hdf5.capture.get() == 1


@pytest.mark.parametrize(
    "scaninfo",
    [
        ({"filepath": "test.h5", "successful": True, "done": False, "scan_id": "123"}),
        ({"filepath": "test.h5", "successful": False, "done": True, "scan_id": "123"}),
        ({"filepath": "test.h5", "successful": None, "done": True, "scan_id": "123"}),
    ],
)
def test_publish_file_location(mock_det, scaninfo):
    mock_det.scaninfo.scan_id = scaninfo["scan_id"]
    mock_det.filepath = scaninfo["filepath"]
    mock_det.custom_prepare.publish_file_location(
        done=scaninfo["done"], successful=scaninfo["successful"]
    )
    if scaninfo["successful"] is None:
        msg = messages.FileMessage(file_path=scaninfo["filepath"], done=scaninfo["done"])
    else:
        msg = messages.FileMessage(
            file_path=scaninfo["filepath"], done=scaninfo["done"], successful=scaninfo["successful"]
        )
    expected_calls = [
        mock.call(
            MessageEndpoints.public_file(scaninfo["scan_id"], mock_det.name),
            msg,
            pipe=mock_det.connector.pipeline.return_value,
        ),
        mock.call(
            MessageEndpoints.file_event(mock_det.name),
            msg,
            pipe=mock_det.connector.pipeline.return_value,
        ),
    ]
    assert mock_det.connector.set_and_publish.call_args_list == expected_calls


@pytest.mark.parametrize("detector_state, expected_exception", [(1, False), (0, True)])
def test_arm_acquisition(mock_det, detector_state, expected_exception):
    with mock.patch.object(mock_det, "stop") as mock_stop:
        mock_det.state._read_pv.mock_data = detector_state
        if expected_exception:
            with pytest.raises(FalconTimeoutError):
                mock_det.timeout = 0.1
                mock_det.custom_prepare.arm_acquisition()
                mock_stop.assert_called_once()
        else:
            mock_det.custom_prepare.arm_acquisition()
            assert mock_det.start_all.get() == 1


def test_trigger(mock_det):
    with mock.patch.object(mock_det.custom_prepare, "on_trigger") as mock_on_trigger:
        mock_det.trigger()
        mock_on_trigger.assert_called_once()


@pytest.mark.parametrize("stopped, expected_abort", [(False, False), (True, True)])
def test_unstage(mock_det, stopped, expected_abort):
    with (
        mock.patch.object(mock_det.custom_prepare, "finished") as mock_finished,
        mock.patch.object(
            mock_det.custom_prepare, "publish_file_location"
        ) as mock_publish_file_location,
    ):
        mock_det.stopped = stopped
        if expected_abort:
            mock_det.unstage()
            assert mock_det.stopped is stopped
            assert mock_publish_file_location.call_count == 0
        else:
            mock_det.unstage()
            mock_finished.assert_called_once()
            mock_publish_file_location.assert_called_with(done=True, successful=True)
            assert mock_det.stopped is stopped


def test_stop(mock_det):
    with (
        mock.patch.object(mock_det.custom_prepare, "stop_detector") as mock_stop_det,
        mock.patch.object(
            mock_det.custom_prepare, "stop_detector_backend"
        ) as mock_stop_detector_backend,
    ):
        mock_det.stop()
        mock_stop_det.assert_called_once()
        mock_stop_detector_backend.assert_called_once()
        assert mock_det.stopped is True


@pytest.mark.parametrize(
    "stopped, scaninfo",
    [
        (False, {"num_points": 500, "frames_per_trigger": 1}),
        (True, {"num_points": 500, "frames_per_trigger": 1}),
    ],
)
def test_finished(mock_det, stopped, scaninfo):
    with (
        mock.patch.object(mock_det.custom_prepare, "stop_detector") as mock_stop_det,
        mock.patch.object(
            mock_det.custom_prepare, "stop_detector_backend"
        ) as mock_stop_file_writer,
    ):
        mock_det.stopped = stopped
        mock_det.dxp.current_pixel._read_pv.mock_data = int(
            scaninfo["num_points"] * scaninfo["frames_per_trigger"]
        )
        mock_det.hdf5.array_counter._read_pv.mock_data = int(
            scaninfo["num_points"] * scaninfo["frames_per_trigger"]
        )
        mock_det.scaninfo.frames_per_trigger = scaninfo["frames_per_trigger"]
        mock_det.scaninfo.num_points = scaninfo["num_points"]
        mock_det.custom_prepare.finished()
        assert mock_det.stopped is stopped
        mock_stop_det.assert_called_once()
        mock_stop_file_writer.assert_called_once()
